/*

Copyright 2010 Andreas Massouras

This file is part of Primer-Initiated Sequence Synthesis (PrInSeS)

For license information, please refer to the LICENSE document included in the distribution.
By downloading or using this software, you accept and are bound by the terms of the license.

*/


#ifndef DNASEQUENCE_H
#define DNASEQUENCE_H

#include <iostream>
#include <limits>
#include "Vector.h"
#include "Util.h"

using namespace std;

typedef enum { _A_ = 0, _C_ = 1, _G_ = 2, _T_ = 3, _N_ = 4 } Base;
inline void operator++(Base& b) { b = (Base) ((int)b+1); }
inline void operator--(Base& b) { b = (Base) ((int)b-1); }
inline Base complement(Base b) { return b == _N_ ? _N_ : (Base) (3 - b); }
static const char mapBase[] = { 'A', 'C', 'G', 'T', 'N' };
inline ostream& operator<<(ostream& o, Base b) { return o << mapBase[b]; }


class DnaSequence {
public:
	static const int64_t maxLength = -(1LL<<62);
	static const GenomicPosition notFound;// = invalidGenomicPosition

	DnaSequence() {}
	DnaSequence(const char *s) : _s(s) {}
	DnaSequence(const char *s, GenomicPosition len) : _s(s, len) {}
	DnaSequence(const String& s) : _s(s) {}
	DnaSequence(const DnaSequence& s) : _s(s._s) {}
	char operator[] (GenomicPosition a) const { return _s[a]; }
	char& operator[] (GenomicPosition a) { return _s[a]; }
	const String& toText() const { return _s; }
	const DnaSequence& operator= (const DnaSequence& s) { _s = s._s; return *this; }
	void set(const DnaSequence& s) { _s = s._s; }
	const DnaSequence& operator= (const String& s) { _s = s; return *this; }
	void set(const String& s) { _s = s; }
	void set(const char* s, GenomicPosition len) { _s.assign(s, len); }
	const DnaSequence& operator= (const char* s) { _s = s; return *this; }
	bool operator< (const DnaSequence& s) const { return _s < s._s; }
	bool operator> (const DnaSequence& s) const { return _s > s._s; }
	bool operator== (const DnaSequence& s) const { return _s == s._s; }
	bool operator== (const char* s) const { return _s == s; }
	bool operator!= (const DnaSequence& s) const { return _s != s._s; }
	const DnaSequence& operator+= (char c) { _s += c; return *this; }
	const DnaSequence& operator+= (Base b) { _s += mapBase[b]; return *this; }
	const DnaSequence& operator+= (const DnaSequence &s) { _s += s._s; return *this; }
	const DnaSequence& operator+= (const String &s) { _s += s; return *this; }
	DnaSequence operator+(char c) const { return DnaSequence(_s + c); }
	DnaSequence operator+(Base b) const { return DnaSequence(_s + mapBase[b]); }
	DnaSequence operator+(const DnaSequence& other) const { return DnaSequence(_s + other._s); }
	int compare(const DnaSequence& b) const { return _s.compare(b._s); }
	int compare(GenomicPosition pos, GenomicPosition n, const char* seq2) const;
	bool equals(const char* s, GenomicPosition n) const { return memcmp(_s.c_str(), s, n) == 0; }
	GenomicPosition length() const { return _s.length(); }
	bool empty() const { return _s.empty(); }
//	operator const String& () const { return _s; }
	void clear() { _s.clear(); }
	DnaSequence substr (int64_t pos, int64_t n = maxLength) const {
		if (pos < 0) pos += _s.length();
		if (n == maxLength) return DnaSequence(_s.substr(pos));
		if (n < 0) n += _s.length() - pos;
		return DnaSequence(_s.substr(pos, n));
	}
	const char* c_str() const { return _s.c_str(); }
	DnaSequence reverseComplement() const { return DnaSequence(Util::reverseComplement(_s)); }

	GenomicPosition find(const DnaSequence& other, int64_t thisFrom = 0, int64_t thisLen = maxLength, int64_t otherFrom = 0, int64_t otherLen = maxLength) const;

	bool contains(const DnaSequence& other, int64_t thisFrom = 0, int64_t thisLen = maxLength, int64_t otherFrom = 0, int64_t otherLen = maxLength) const 
	{
		return find(other, thisFrom, thisLen, otherFrom, otherLen) != notFound;
	}

	// The following are equivalent to the above find(), but are faster if you perform multiple calls to find() on the same target string
	// First create a PreparedFindTarget with the target DnaSequence (aka the "needle")
	class PreparedFindTarget {
	private:
		GenomicPosition _badCharShift[256];

	public:
		const char* _needle;
		int64_t _nlen;
		PreparedFindTarget(const DnaSequence&, int64_t from, int64_t length);
		bool valid() const { return _needle != NULL; }
		GenomicPosition badCharShift(char c) const { return _badCharShift[(int)c]; }
	};
	PreparedFindTarget findPrepare(int64_t from, int64_t length) const;
	// then run this (many times) on the longer DnaSequence (aka "haystack"):
	GenomicPosition find(const PreparedFindTarget& other, int64_t thisFrom = 0, int64_t thisLen = maxLength) const;

	void resize(int64_t n) { if (n < 0) n += _s.length(); _s.resize(n); }

private:
	String _s;
};

inline ostream& operator<<(ostream& o, const DnaSequence& s)
{
	return o << s.toText();
}

class DnaNamedSequence : public DnaSequence {
public:
	DnaNamedSequence() {}
	DnaNamedSequence(const char* name, const char* seq) : DnaSequence(seq), _name(name) {}
	DnaNamedSequence(const char* name, const char* seq, GenomicPosition len) : DnaSequence(seq, len), _name(name) {}
	DnaNamedSequence(const String& name, const String& seq) : DnaSequence(seq), _name(name) {}
	DnaNamedSequence(const DnaNamedSequence& s) : DnaSequence(s), _name(s._name) {}
	DnaNamedSequence(const DnaSequence& s, const String& name) : DnaSequence(s), _name(name) {}
	DnaNamedSequence(const DnaSequence& s) : DnaSequence(s) {}
	const String& name() const { return _name; }
	void setName(const String& n) { _name = n; }
	void setName(const char* n) { _name = n; }
private:
	String _name;
};

inline ostream& operator<<(ostream& o, const DnaNamedSequence& s) { Util::printFasta(o, s.name(), s.toText()); return o; }

void loadFastaFile(const char* filename, /* returned: */ Vector<DnaNamedSequence*>&);

// this is used to sort lists of sequences by name:
int nameBefore(const DnaNamedSequence&, const DnaNamedSequence&);

#endif
