/*

Copyright 2010 Andreas Massouras

This file is part of Primer-Initiated Sequence Synthesis (PrInSeS)

For license information, please refer to the LICENSE document included in the distribution.
By downloading or using this software, you accept and are bound by the terms of the license.

*/

#include "Prinses.h"
#include <fstream>
#include <queue>
#include <map>
#include <iomanip>
#include <assert.h>

#define PRINSES_DEBUG

namespace Prinses {

// typedef enum { END_TERMINATOR_FOUND, END_LOW_COVER, END_MAX_LENGTH, END_REPEATS, END_GENOME_REPEAT } AssemblyEndReason;
const char assemblyEndReasonChar[] = { 'f', 'l', 'm', 'r', 'g' };

class AssemblyQueueItem
{
public:
	DnaSequence _seq;
	int _numBranches;
	uint32_t _score;
	AssemblyQueueItem(const DnaSequence& seq, int numBranches, uint32_t score) : _seq(seq), _numBranches(numBranches), _score(score) {}
	AssemblyQueueItem() : _numBranches(0), _score(0) {}
};


AssemblyParameters::AssemblyParameters(ostream* viewFile,
	bool verbose, bool useScores, int maxBranches, const ShortSequenceSortedArray& reads, 
	uint16_t minDepth, uint16_t maxDepth,
	uint16_t minMatchLength, GenomicPosition maxAssemblyLength, uint16_t numTerminators, uint16_t terminatorLength,
	const GenomeFragmentArray& pos, const GenomeFragmentArray& neg)
		: _viewFile(viewFile), _verbose(verbose), _useScores(useScores),
		_maxBranches(maxBranches),
		_reads(reads), _minDepth(minDepth),
		_maxDepth(maxDepth), _minMatchLength(minMatchLength), _maxAssemblyLength(maxAssemblyLength),
		_terminators(numTerminators, NULL), _terminatorLength(terminatorLength),
		_genomeFragmentsPositive(pos), _genomeFragmentsNegative(neg)
{
}

class ReadUsed
{
public:
	SequencingRead _read;
	int _overlap;
	int _score;
	ReadUsed(const SequencingRead& r, int o, int s) : _read(r), _overlap(o), _score(s) {}
};


void
assemble(AssemblyParameters& params, list<AssemblyResult>& results)
{
	results.clear();
	const int readLengthM1 = params._reads.readLength() - 1;	// this must be signed
	const uint16_t primerLength = params._primer.length();
	if (primerLength < 4) throw(1);
	Stack<AssemblyQueueItem> queue;
	queue.push(AssemblyQueueItem(params._primer, 0, 0));
	bool verbose = params._verbose and params._viewFile;
	SequencingRead read;
	uint8_t fragBuff[ShortDnaBinarySequence::memorySize(readLengthM1)];
	ShortDnaBinarySequence* frag = (ShortDnaBinarySequence*) fragBuff;
	const uint16_t termLen = params._terminatorLength;

	if (readLengthM1 < 5 or readLengthM1 > 100000) {
		cerr << "Prinses: read length " << (readLengthM1+1) << " out of range\n";
		exit(1);
	}

	if (params._minMatchLength < 16) params._minMatchLength = 16;

	while (!queue.isEmpty()) {
		AssemblyQueueItem qitem = queue.pop();
		GenomicPosition seqlen = qitem._seq.length();

/* Check if in doneQueue - to be implemented
		if (qitem._seq.length() >= readLengthM1) {
			
		}
*/
		AssemblyResult result(qitem._seq, END_MAX_LENGTH, qitem._score);
		ShortDnaSequenceTree subSequences;
		for (GenomicPosition position = 0; position + readLengthM1 <= seqlen; position++) {
			subSequences.addNode(result._contig.substr(position, readLengthM1).c_str());
		}

		// main loop to add one base at a time
		for (GenomicPosition position = seqlen + 1; position <= params._maxAssemblyLength; ++position) {
			seqlen = result._contig.length();
			int numreads = 0;
			int thisScore[5] = { 0, 0, 0, 0, 0};
			int numreadsForBase[5] = { 0, 0, 0, 0, 0 };
			int totalScore = 0;
			vector<ReadUsed> readsUsed;

			// loops once for each size of overlap
			for (GenomicPosition i = seqlen > (GenomicPosition) readLengthM1 ? readLengthM1 : seqlen; i >= params._minMatchLength; --i) {
				if (params._maxDepth and numreads >= params._maxDepth) break;
				if (!frag->set(result._contig.c_str() + seqlen - i)) continue;
				ShortSequenceSortedArrayIterator readsIter(params._reads.findFragment(frag));
//				cdbg << "ReadsIter=" << readsIter._current << "-" << readsIter._last << endl;
				while (readsIter.next(read)) {
					Base base = read[i];
					int num = read.count();
					int score = 0;
					if (params._useScores) {
						score = read.score(i);
						thisScore[base] += score * num;
						totalScore += score * num;
					}
					numreads += num;
					numreadsForBase[base] += num;
					if (verbose)
						readsUsed.push_back(ReadUsed(read, i, score));
				}
			}
			if (verbose) {
				*params._viewFile << result._contig.substr(-readLengthM1) << " " << params._assemblyId << ":" << position << endl;
				vector<ReadUsed>::iterator it;
				for (it = readsUsed.begin(); it != readsUsed.end(); it++) {
					if (readLengthM1 > it->_overlap)
						*params._viewFile << String(readLengthM1 - it->_overlap, ' ');
					*params._viewFile << (*it)._read.sequence() << " x+" << (*it)._read.count() << " " << it->_score << endl;
				}
			}

			if (numreads < params._minDepth) {
				result._assemblyEndReason = END_LOW_COVER;
				break;
			}

			// Call the next base to be added to the assembled sequence:
			Base callBase = _N_;
			int max = 0;
			Base i;
			for (i = _A_; i < _N_; ++i) {
				if (max < numreadsForBase[i]) {
					callBase = i;
					max = numreadsForBase[i];
				}
			}
			if (max * 5 < numreads * 4) {		// if less than 80% of reads agree

				// Add branch to find heterozygous variants:
				if (params._maxBranches and qitem._numBranches < params._maxBranches) {
					if (params._useScores) {
						max = 0;
						Base callBase2 = _N_;
						int max2 = 0;
						for (i = _A_; i < _N_; ++i) {
							if (max < thisScore[i]) {
								callBase2 = callBase;
								max2 = max;
								callBase = i;
								max = thisScore[i];
							} else if (max2 < thisScore[i]) {
								callBase2 = i;
								max2 = thisScore[i];
							}
						}
						// if second call accounts for more than 1/3 of score-weighted number of reads:
						if (max2 * 3 >= totalScore) {
							queue.push(AssemblyQueueItem(result._contig + callBase2, qitem._numBranches + 1, result._score + max2));
						}
					} else {
						Base callBase2 = _N_;
						int max2 = 0;
						for (i = _A_; i < _N_; ++i) {
							if (i == callBase) continue;
							if (max2 < numreadsForBase[i]) {
								callBase2 = i;
								max2 = numreadsForBase[i];
							}
						}
						// if second call accounts for more than 1/3 of reads:
						if (max2 * 3 > numreads) {
							queue.push(AssemblyQueueItem(result._contig + callBase2, qitem._numBranches + 1, result._score + max2));
						}
					}

				// Use score-weighted majority
				} else if (params._useScores) {
					max = 0;
					for (i = _A_; i < _N_; ++i) {
						if (max < thisScore[i]) {
							callBase = i;
							max = thisScore[i];
						}
					}
				}
			}

			if (verbose) {
				*params._viewFile << params._assemblyId << "\t" << position << "\t" << totalScore << "\t" << numreads << "\t" << callBase;
				if (params._useScores and totalScore) {
					Base i;
					for (i = _A_; i < _N_; ++i) {
						if (!thisScore[i]) continue;
						*params._viewFile << "\t" << i << ":" << (thisScore[i] * 100 / totalScore) << "%:" << numreadsForBase[i];
					}
				}
				*params._viewFile << endl;
			}

			result._contig += callBase;
			result._score += params._useScores ? thisScore[callBase] : numreadsForBase[callBase];
			seqlen++;


			// see if sequence is repeating:
			if (seqlen > (GenomicPosition) readLengthM1 and 1 < subSequences.addNode(result._contig.substr(-readLengthM1).c_str())) {
				result._assemblyEndReason = END_REPEATS;
				break;
			}

			// see if we have reached a terminator:
			const char* term = result._contig.c_str() + seqlen - termLen;
			const Vector<const char*>& T = params._terminators;
			for (uint16_t i = 0; i < T.size(); ++i) {
				if (T[i] == NULL) continue;
				if (position >= (GenomicPosition) termLen + primerLength and strncmp(term, T[i], termLen) == 0) {
					result._contig.resize(-(int64_t) params._terminatorLength);
					result._assemblyEndReason = END_TERMINATOR_FOUND;
					result._terminatorIndex = i;
					break;
				}
			}
			if (result._assemblyEndReason == END_TERMINATOR_FOUND) break;

			// see if we have reached a repeat on the genome:
/*
			GenomeFragmentArrayIterator it(params._genomeFragments.findFragment(result._contig.c_str(), primerLength));
			if (it.numberRemaining() > 1) {
				result._assemblyEndReason = END_GENOME_REPEAT;
				break;
			}
*/
		}

		result._contig = result._contig.substr(primerLength); // get rid of primer
		results.push_back(result);
	}
}


uint16_t	// returns average read depth
loadMaqProfile(ifstream& ifile, Vector<DnaNamedSequence*>& maqConsensus, map<String, const uint8_t*>* maqReadDepth)
{
/* fields:
$seqName, $position, $referenceBase, 
$consensusBase, $quality, $readDepth, $avgNumberHits,
$maxQuality, $minConsensusFlanking6, $secondBest, 
$logLikelihoodSecondBest
NC_001422.1	35	C	C	201	146	1.00	63	62	N	255	N
*/
	// maq view file fields:
	string seqName, position, referenceBase, consensusBase, quality, readDepth, avgNumberHits, maxQuality, minConsensusFlanking6, secondBest, logLikelihoodSecondBest;

	list<DnaNamedSequence*> maqConsensusList;
	if (maqReadDepth) maqReadDepth->clear();

	ifile.seekg(0, ios::end);
	streampos length = ifile.tellg();
	ifile.seekg(0, ios::beg);
	char* buffer = new char[length + (streampos) 1];
	ifile.read(buffer, length);
	buffer[length] = 0;
	cerr << Util::getLocalTime(time(NULL)) << " maq file size " << length << " bytes\n";
	map<string, GenomicPosition> seqLength;
	string *fields1[] = { &seqName, NULL };

	uint32_t totalReadDepth = 0, i;
	const char* str = buffer;
	string currentSeqName;
	GenomicPosition* currentSeqLength = NULL;
	for (i = 0; Util::parseTabSeparatedFile(str, fields1); ++i) {
		if (i % 1000000 == 0) {
			cerr << "\rloading maq view file " << setw(4) << (i/1000000) << "m lines";
		}
		if (seqName != currentSeqName) {
			currentSeqLength = &(seqLength[seqName] = 1);
			currentSeqName = seqName;
		} else {
			(*currentSeqLength)++;
		}

	}

	str = buffer;
	string *fields2[] = {
		&seqName, &position, &referenceBase, &consensusBase, &quality, &readDepth, &avgNumberHits, &maxQuality, /* &minConsensusFlanking6, &secondBest, &logLikelihoodSecondBest, */ NULL };
	currentSeqName.clear();
	char *consensusRaw = NULL;
	uint8_t *readDepthRaw = NULL;
	int numf;
	for (i = 0; (numf = Util::parseTabSeparatedFile(str, fields2)); ++i) {
		if (i % 1000000 == 0) {
			cerr << "\rloading maq view file " << setw(4) << (i/1000000) << "m lines";
		}
		
		if (numf < 6) { 
			cerr << "line " << i << " not found expected fields\n";
			throw(1);
		}
		if (seqName != currentSeqName) {
			if (consensusRaw) {
				maqConsensusList.push_back(new DnaNamedSequence(currentSeqName.c_str(), consensusRaw));
				delete [] consensusRaw;
				if (maqReadDepth)
					(*maqReadDepth)[currentSeqName] = readDepthRaw;
			}
			GenomicPosition len = seqLength[seqName];
			consensusRaw = new char[len+1];
			consensusRaw[len] = 0;
			if (maqReadDepth)
				readDepthRaw = new uint8_t[len];
			currentSeqName = seqName;
		}
		GenomicPosition pos;
		if (!Util::value<GenomicPosition>(pos, position)) {
			cerr << "line " << i << " expected integer for position field\n";
			throw(2);
		}
		pos--;
		
		switch(consensusBase[0]) {
		case 'A':
		case 'C':
		case 'G':
		case 'T':
			consensusRaw[pos] = consensusBase[0];
			break;
		default:
			consensusRaw[pos] = referenceBase[0];
		}
		GenomicPosition d;
		if (!Util::value<GenomicPosition>(d, readDepth)) {
			cerr << "line " << i << " expected integer for read depth field\n";
			throw(3);
		}
#if 0
		int32_t maxQ;
		if (!Util::value<int32_t>(maxQ, maxQuality)) {
			cerr << "line " << i << " expected integer for max quality field\n";
			throw(4);
		}
		if (maxQ < 1) d = 0;
#endif
		if (readDepthRaw) readDepthRaw[pos] = d;
		totalReadDepth += d;
	}
	if (consensusRaw) {
		maqConsensusList.push_back(new DnaNamedSequence(currentSeqName.c_str(), consensusRaw));
		delete [] consensusRaw;
		if (maqReadDepth)
			(*maqReadDepth)[currentSeqName] = readDepthRaw;
	}
	uint16_t averageReadDepth = i ? (totalReadDepth / i) : 0;
	cerr << "\rloaded maq view file (" << i << ") lines, average depth " << averageReadDepth << endl;
	delete [] buffer;
	maqConsensus.resize(maqConsensusList.size());
	list<DnaNamedSequence*>::iterator it;
	ChromosomeIndex ci;
	for (it = maqConsensusList.begin(), ci = 0; it != maqConsensusList.end(); ++it, ++ci)
		maqConsensus[ci] = *it;
	return averageReadDepth;
}


uint64_t
mapReads(const char* refseq, const GenomicPosition refseqLength, const ShortSequenceSortedArray& reads, const GenomeFragmentArray& genomeFragments, 
	uint16_t minMatchLength, /* returned: */ float*& coverage, ostream* logFile)
{
	if (!refseqLength) {
		if (logFile) *logFile << "Empty sequence!\n";
		coverage = NULL;
		return 0;
	}
	const uint16_t readLength = reads.readLength();
	if (minMatchLength > readLength) minMatchLength = readLength;
	uint8_t fragBuff[ShortDnaBinarySequence::memorySize(readLength)];
	ShortDnaBinarySequence* frag = (ShortDnaBinarySequence*) fragBuff;

	coverage = new float[refseqLength];
	for (GenomicPosition pos = 0; pos < refseqLength; pos++)
		coverage[pos] = 0.0;
	SequencingRead read;
	float ret = 0.0;

	for (GenomicPosition pos = 0; pos + minMatchLength <= refseqLength; pos++) {
		if (logFile and (pos+1) % 1000000 == 0) *logFile << '.';
		const uint16_t fragLen = pos + readLength <= refseqLength ? readLength : refseqLength - pos;
		if (!frag->set(refseq + pos, fragLen)) continue;
		ShortSequenceSortedArrayIterator readsIter(reads.findFragment(frag, minMatchLength));
		while (readsIter.next(read)) {
			uint16_t len = read.length();
			if (!frag->equals(read.sequence(), len)) continue;

			uint32_t numCopies = read.count();
			GenomeFragmentArrayIterator fit(genomeFragments.findFragment(refseq + pos, len));
			float cover = numCopies;
			uint32_t numGenomeFragments = fit.numberRemaining();
			if (numGenomeFragments > 1) {
				cover /= (float) numGenomeFragments;
			}
			ret += cover;
			for (GenomicPosition p = pos; len; ++p, --len) {
				coverage[p] += cover;
			}
		}
	}

	return (uint64_t) ret;
}

static const GenomicPosition kmerSize = 11;
static const uint16_t maxAlignments = 256;

int
mapSingleRead(const char* readseq, const GenomicPosition length, const GenomeFragmentArray& genomeFragments, /* returned: */ Alignment& alignment)
{
	assert(length >= kmerSize);

	//cdbg << "looking for " << String(readseq, length) << endl;
	Alignment alignments[maxAlignments];
	uint32_t numFound = 0;
	const Genome& genome(genomeFragments.genome());
	const char* const genomeSequence = genome.sequence();
	char reads[2][length+1];
	for (GenomicPosition i = 0, j = length - 1; i < length; ++i, --j) {
		switch(readseq[i]) {
		case 'A': case 'a':
			reads[0][i] = 'A';
			reads[1][j] = 'T';
			break;
		case 'C': case 'c':
			reads[0][i] = 'C';
			reads[1][j] = 'G';
			break;
		case 'G': case 'g':
			reads[0][i] = 'G';
			reads[1][j] = 'C';
			break;
		case 'T': case 't':
			reads[0][i] = 'T';
			reads[1][j] = 'A';
			break;
		default:		// make any other character (incl N) unmatchable to the genome fragments
			reads[0][i] = '.';
			reads[1][j] = '.';
		}
	}
	reads[0][length] = reads[1][length] = 0;
	uint16_t bestNumMismatches = length / kmerSize - 1;
	uint16_t numberHitsAtBest = 0;
	uint32_t numberOfKmersChecked = 3;
	if (numberOfKmersChecked > length / kmerSize) numberOfKmersChecked = length / kmerSize;

	for (uint32_t kmerNumber = 0; numFound < maxAlignments and kmerNumber < numberOfKmersChecked; kmerNumber++) {
		GenomicPosition offset = kmerNumber * kmerSize;
		for (int negStrand = 0; numFound < maxAlignments and negStrand < 2; ++negStrand) {
			const char* const read = reads[negStrand];
			GenomeFragmentArrayIterator git(genomeFragments.findFragment(read + offset, kmerSize));
			GenomicPosition pos;
			while (numFound < maxAlignments and git.next(pos)) {
				uint32_t thisNumMismatches = 0;
				if (kmerNumber) {
					if (pos < offset) continue;
					for (GenomicPosition gp = pos - offset, i = 0; gp < pos; gp++, i++) {
						if (genomeSequence[gp] != read[i]) {
							thisNumMismatches++;
							if (thisNumMismatches > bestNumMismatches) break;
						}
					}
					if (thisNumMismatches > bestNumMismatches) continue;
				}
				for (GenomicPosition gp = pos + kmerSize, i = offset + kmerSize; i < length; gp++, i++) {
					if (genomeSequence[gp] != read[i]) {
						thisNumMismatches++;
						if (thisNumMismatches > bestNumMismatches) break;
					}
				}
				if (thisNumMismatches > bestNumMismatches) continue;
				if (thisNumMismatches < bestNumMismatches) {
					bestNumMismatches = thisNumMismatches;
					numberHitsAtBest = 1;
				} else {
					numberHitsAtBest++;
					// get out if more than one perfect hits:
					if (numberHitsAtBest > 1 and bestNumMismatches == 0)
						return -1;
				}
				alignments[numFound]._position = pos - offset;
				alignments[numFound]._negStrand = negStrand;
				alignments[numFound]._numMismatches = thisNumMismatches;
				numFound++;
			}
		}
		if (bestNumMismatches == kmerNumber and numFound < maxAlignments and numberHitsAtBest == 1) {
			for (uint16_t i = 0; i < numFound; ++i) {
				if (alignments[i]._numMismatches == bestNumMismatches) {
					alignment = alignments[i];
					return 1;
				}
			}
			abort(); // shouldn't get here
		}
	}

	if (numFound == maxAlignments) return -1;

	bool found = false;
	uint16_t foundIndex = 0;
	for (uint16_t i = 0; i < numFound; ++i) {
		if (alignments[i]._numMismatches == bestNumMismatches) {
//			if (n) cdbg << genome.toText(alignmentPositions[i].position()) << '\t' << alignmentPositions[i].strand() << '\t' << n << endl;
			if (found) return -1;
			found = true;
			foundIndex = i;
		}
	}
	if (found) {
		alignment = alignments[foundIndex];
		return 1;
	}
	return 0;
}


}; // namespace Prinses
