/*

Copyright 2011 Andreas Massouras

This file is part of Primer-Initiated Sequence Synthesis (PrInSeS)

For license information, please refer to the LICENSE document included in the distribution.
By downloading or using this software, you accept and are bound by the terms of the license.

*/

#include "SamPileup.h"


static Base samBaseToBaseMap[16] = { _N_, _A_, _C_, _N_, _G_, _N_, _N_, _N_, _T_, _N_, _N_, _N_, _N_, _N_, _N_, _N_ };

int
SamPileup::pileupCallback(uint32_t chromoId, uint32_t chromoPos, int readDepth, const bam_pileup1_t *pl, void *data)
{
	SamPileup* sam = (SamPileup*) data;
	sam->pileupCallback(chromoId, chromoPos, readDepth, pl);
	return 0;
}


SamPileup::SamPileup(const char* filename, const Genome& genome, uint16_t minDepth, uint8_t minQuality)
	: _referenceGenome(genome), _targetGenome(genome), _snpCount(0), _minDepth(minDepth), _minQuality(minQuality)
{
	cerr << Util::getLocalTime(time(NULL)) << " loading BAM file " << filename << endl;
	samfile_t *samfile = samopen(filename, "rb", 0);
	if (samfile == 0) {
		cerr << "Can't open BAM file " << filename << endl;
		exit(1);
	}
	int32_t nTargets = samfile->header->n_targets;
	_samChromosomeMap.resize(nTargets);
	bool failed = false;
	for (int32_t i = 0; i < nTargets; ++i) {
		ChromosomeIndex ci = genome.chromoIndex(samfile->header->target_name[i]);
		if (ci == invalidChromosomeIndex) {
			cerr << "Unrecognised chromosome " << samfile->header->target_name[i] << endl;
			failed = true;
			continue;
		}
		if (samfile->header->target_len[i] != genome.chromoLength(ci)) {
			cerr << "Different chromosome lengths between BAM file and genome for chromosome " << genome.chromoName(ci)
				 << " (" << samfile->header->target_len[i] << " vs " << genome.chromoLength(ci) << ')' << endl;
			failed = true;
			continue;
		}
		_samChromosomeMap[i] = ci;
	}
	if (failed) exit(1);

	_genomeReadDepth.resize(genome.length());
	_genomeReadDepth.set(0);
	sampileup(samfile, -1, pileupCallback, this);
	samclose(samfile);
	cerr << Util::getLocalTime(time(NULL)) << " BAM file loaded, SNP count " << _snpCount << " average depth " << averageReadDepth() << endl;
}


// SNP calling function
void
SamPileup::pileupCallback(uint32_t chromoId, uint32_t chromoPos, int readDepth, const bam_pileup1_t* pl)
{
	uint32_t baseCount[4] = { 0,0,0,0 };
	uint32_t baseSumQuality[4] = { 0,0,0,0 };
	uint32_t numReadsUsed = 0;
	for (int r = 0; r < readDepth; ++r) {
		const bam_pileup1_t& al = pl[r];

		if (al.b->core.qual < _minQuality) continue;

		// 1=A, 2=C, 4=G, 8=T, 15=N
		uint8_t callBase = bam1_seqi(bam1_seq(al.b), al.qpos);
		assert(callBase == 1 or callBase == 2 or callBase == 4 or callBase == 8 or callBase == 15);
		if (callBase == 15) continue;

		// skip if not everything in cigar string is a (mis)match:
		uint32_t nCigar = al.b->core.n_cigar;
		if (nCigar != 1) continue;
		uint32_t* cigar = bam1_cigar(al.b);
		if ((cigar[0] & BAM_CIGAR_MASK) != BAM_CMATCH) continue;

		Base cb = samBaseToBaseMap[callBase];
		baseCount[cb] ++;
		baseSumQuality[cb] += al.b->core.qual;
		numReadsUsed++;

//		const uint32_t readLength = al.b->core.l_qseq;
/*
		for (uint32_t k = 0; k < nCigar; ++k) {
			int cop =cigar[k] & BAM_CIGAR_MASK; // operation
			int cl = cigar[k] >> BAM_CIGAR_SHIFT; // length
			switch(cop) {
			 case BAM_CMATCH: cdbg << "M";break;
			 case BAM_CINS: cdbg << ("I");break;
			 case BAM_CDEL: cdbg << ("D");break;
			 case BAM_CREF_SKIP: cdbg << ("N"); break;
			 case BAM_CSOFT_CLIP: cdbg << ("S");break;
			 case BAM_CHARD_CLIP: cdbg << ("R");break;
			 case BAM_CPAD: cdbg << ("P");break;
			 default:cdbg << ("?");break;
			 }
			cdbg << cl;
		}
*/
	}
	if (numReadsUsed < _minDepth) return;

	const GenomicPosition pos = _referenceGenome.position(_samChromosomeMap[chromoId], chromoPos);
	if (pos == invalidGenomicPosition) return;
	uint32_t maxSumQuality = 0;
	Base callBase = _N_;
//	_referenceGenome.output(cdbg, pos);
	char refBase = *_referenceGenome.sequence(pos);
	uint32_t numBases = 0;
	for (Base i = _A_; i < _N_; ++i) {
//		cdbg << ' ' << mapBase[i] << ':' << baseCount[i];
		if (baseCount[i]) numBases++;
		if (baseCount[i] < _minDepth) continue;
		if (baseSumQuality[i] > maxSumQuality) {
			maxSumQuality = baseSumQuality[i];
			callBase = i;

		// bias ties towards reference base:
		} else if (baseSumQuality[i] == maxSumQuality and mapBase[i] == refBase) {
			callBase = i;
		}
	}

	_genomeReadDepth[pos] = numReadsUsed;
	if (numBases >= 3) return;
	if (callBase == _N_) return;

	// any SNP call must be supported by 80% of all reads
	if (baseCount[callBase] * 5 < numReadsUsed * 4) return;

	if (refBase != mapBase[callBase]) {
#if 0
		_referenceGenome.output(cdbg, pos);
		cdbg << " ref=" << *(_referenceGenome.sequence(pos)) << " call=" << callBase;
		for (Base i = _A_; i < _N_; ++i) {
			if (baseCount[i])
				cdbg << ' ' << mapBase[i] << ':' << baseCount[i] << ':' << baseSumQuality[i];
		}
		cdbg << endl;
#endif
		_snpCount++;
		_targetGenome.setBase(pos, callBase);
	}
}


double
SamPileup::averageReadDepth(GenomicPosition pos, GenomicPosition len, GenomicPosition flankingLen) const
{
	const ChromosomeIndex ci = _referenceGenome.chromoIndex(pos);
	const GenomicPosition chromoStart = _referenceGenome.chromoStart(ci);
	const GenomicPosition chromoLength = _referenceGenome.chromoLength(ci);
	const GenomicPosition chromoPos = pos - chromoStart;
	if (chromoPos + flankingLen + len >= chromoLength) {
		len = chromoLength - chromoPos;
	} else {
		len += flankingLen;
	}
	if (chromoPos < flankingLen) {
		len += flankingLen - chromoPos;
		pos = chromoStart;
	} else {
		len += flankingLen;
		pos -= flankingLen;
	}
	if (len == 0) return 0.0;
	uint64_t sum = 0;
	for (GenomicPosition l = len; l; l--, pos++) sum += _genomeReadDepth[pos];
	return (double) sum / (double) len;
}


void
SamPileup::outputDepthStats(ostream& o) const
{
	const uint32_t maxDepth = 2000;
	const uint32_t rollingWindowSize = 100;
	VectorI depthBuckets(maxDepth+1, 0);
	GenomicPosition p;
	uint32_t rollingSum = 0;
	for (p = 0; p < rollingWindowSize and p < _genomeReadDepth.size(); p++) {
		rollingSum += _genomeReadDepth[p];
	}
	uint32_t d = rollingSum / rollingWindowSize;
	if (d > maxDepth) d = maxDepth;
	depthBuckets[d]++;

	for (; p < _genomeReadDepth.size(); p++) {
		rollingSum += _genomeReadDepth[p] - _genomeReadDepth[p - rollingWindowSize];
		d = rollingSum / rollingWindowSize;
		if (d >= maxDepth) d = maxDepth;
		depthBuckets[d]++;
	}

	o << "coverage\tnumber_of_" << rollingWindowSize << "bp_windows\tpct" << endl;
	const double denom = _genomeReadDepth.size() - rollingWindowSize + 1;
	if (denom == 0) return;
	for (p = 0; p < maxDepth; p++) {
		if (depthBuckets[p])
			o << p << '\t' << depthBuckets[p] << '\t' << 100.0 * (double) depthBuckets[p] / denom << endl;
	}
	if (depthBuckets[maxDepth])
		o << maxDepth << "+\t" << depthBuckets[maxDepth] << '\t' << 100.0 * (double) depthBuckets[maxDepth] / denom << endl;
}
