/*

Copyright 2010 Andreas Massouras

This file is part of Primer-Initiated Sequence Synthesis (PrInSeS)

For license information, please refer to the LICENSE document included in the distribution.
By downloading or using this software, you accept and are bound by the terms of the license.

*/

#include "ShortDnaBinarySequence.h"

void
ShortDnaBinaryBaseSequence::copy(const ShortDnaBinaryBaseSequence& from, uint16_t n)
{
	uint16_t bytesLeft = n / 4;
	uint8_t* to = _sequence;
	const uint8_t* fr = from._sequence;
	while (bytesLeft) {
		*to = *fr;
		to++;
		fr++;
		bytesLeft--;
	}
	switch(n % 4) {
		case 0:
			break;
		case 1:
			*to = *fr & 0xc0;
			break;
		case 2:
			*to = *fr & 0xf0;
			break;
		case 3:
			*to = *fr & 0xfc;
			break;
	}
}


uint16_t
ShortDnaBinaryBaseSequence::set(const char* cseq)
{
	uint8_t* seq = _sequence;
	uint8_t c = 0;
	uint16_t l = 0;
	while (*cseq) {
		c <<= 2;
		switch((*cseq) & 0xdf) {	// upper case
		case 'A': break;
		case 'C': c ++; break;
		case 'G': c += 2; break;
		case 'T': c += 3; break;
		default:
			return 0;
		}

		l++;
		if (l % 4 == 0) {
			*seq = c;
			seq++;
			c = 0;
		}
		cseq++;
	}
	if (l % 4) {
		c <<= (2 * (4-l%4));
		*seq = c;
	}
	return l;
}


uint16_t
ShortDnaBinaryBaseSequence::set(const char* cseq, uint16_t n)
{
	uint8_t* seq = _sequence;
	uint8_t c = 0;
	uint16_t l = 0;
	while (n) {
		c <<= 2;
		switch((*cseq) & 0xdf) {	// upper case
		case 'A': break;
		case 'C': c ++; break;
		case 'G': c += 2; break;
		case 'T': c += 3; break;
		default:
			return 0;
		}

		l++;
		if (l % 4 == 0) {
			*seq = c;
			seq++;
			c = 0;
		}
		n--;
		cseq++;
	}
	if (l % 4) {
		c <<= (2 * (4-l%4));
		*seq = c;
	}
	return l;
}


int
ShortDnaBinaryBaseSequence::compare(const ShortDnaBinaryBaseSequence& other, uint16_t n) const
{
	uint16_t n8 = n / 4;
	const uint8_t* p1 = _sequence;
	const uint8_t* p2 = other._sequence;
	while (n8) {
		if (*p1 < *p2) return -1;
		if (*p1 > *p2) return 1;
		n8--;
		p1++;
		p2++;
	}
	switch(n % 4) {
	case 0: break;
	case 1:
		if ((*p1 & 0xc0) < (*p2 & 0xc0)) return -1;
		if ((*p1 & 0xc0) > (*p2 & 0xc0)) return 1;
		break;
	case 2:
		if ((*p1 & 0xf0) < (*p2 & 0xf0)) return -1;
		if ((*p1 & 0xf0) > (*p2 & 0xf0)) return 1;
		break;
	case 3:
		if ((*p1 & 0xfc) < (*p2 & 0xfc)) return -1;
		if ((*p1 & 0xfc) > (*p2 & 0xfc)) return 1;
		break;
	}
	return 0;
}


bool
ShortDnaBinaryBaseSequence::equals(const ShortDnaBinaryBaseSequence& other, uint16_t n) const
{
	uint16_t n8 = n / 4;
	const uint8_t* pp1 = _sequence;
	const uint8_t* pp2 = other._sequence;
	while (n8) {
		if (*pp1 != *pp2) return false;
		n8--;
		pp1++;
		pp2++;
	}

	switch(n % 4) {
	case 0:
		break;
	case 1:
		if ((*pp1 & 0xc0) != (*pp2 & 0xc0)) return false;
		break;
	case 2:
		if ((*pp1 & 0xf0) != (*pp2 & 0xf0)) return false;
		break;
	case 3:
		if ((*pp1 & 0xfc) != (*pp2 & 0xfc)) return false;
		break;
	}
	return true;
}


String
ShortDnaBinarySequence::toText() const
{
	String o;
	for (uint16_t i = 0; i < _length; ++i)
		o += mapBase[_baseSeq[i]];
	return o;
}


uint32_t
ShortDnaSequenceTree::addNode(const char* sequence)
{
	_number++;
	ShortDnaSequenceNode* newNode = new (strlen(sequence)) ShortDnaSequenceNode(sequence);
	
	ShortDnaSequenceNode** nodeP = &_root;
	uint64_t depth = 0;

	while (*nodeP) {
		int cmp = newNode->_seq.compare(&((*nodeP)->_seq));
		if (cmp < 0) {
			nodeP = &((*nodeP)->_left);
			depth++;
		} else if (cmp > 0) { 
			nodeP = &((*nodeP)->_right);
			depth++;
		} else {
			delete newNode;
			(*nodeP)->_numCopies++;
			return (*nodeP)->_numCopies;
		}
	}

	*nodeP = newNode;
	_numUnique++;
	depth++;
	if (depth > _maxDepth)
		_maxDepth = depth;
	return 1;
}


bool
ShortDnaSequenceTree::contains(const char* sequence) const
{
	uint8_t buf[ShortDnaBinarySequence::memorySize(strlen(sequence))];
	ShortDnaBinarySequence* seq = (ShortDnaBinarySequence*) buf;
	seq->set(sequence);

	ShortDnaSequenceNode* node = _root;

	while (node) {
		int cmp = seq->compare(&(node->_seq));
		if (cmp < 0) {
			node = node->_left;
		} else if (cmp > 0) { 
			node = node->_right;
		} else {
			return true;
		}
	}

	return false;
}


ShortDnaSequenceTree::~ShortDnaSequenceTree()
{
	Stack<ShortDnaSequenceNode *> stack;
	ShortDnaSequenceNode* node = _root;
	while (node) {
		if (node->_left) {
			stack.push(node);
			node = node->_left;
		} else {
			ShortDnaSequenceNode* r = node->_right;
			delete node;
			node = r;
		}
		while (!node and !stack.isEmpty()) {
			node = stack.pop();
			ShortDnaSequenceNode* r = node->_right;
			delete node;
			node = r;
		}
	}
}
