/*

Copyright 2010 Andreas Massouras

This file is part of Primer-Initiated Sequence Synthesis (PrInSeS)

For license information, please refer to the LICENSE document included in the distribution.
By downloading or using this software, you accept and are bound by the terms of the license.

*/

#ifndef VARIANT_H
#define VARIANT_H

#include <map>
#include <list>
#include "DnaSequence.h"
#include "Genome.h"
//#include "GenomeFragmentArray.h"

class Variant;
typedef list<Variant> VariantSet;	// set of variants for a genome

class Variant
{
public:
	char _direction;//qqqqqq

protected:
	GenomicRegion _region;
	DnaSequence _newSeq;
	String _text;

public:
	Variant() {}
	Variant(GenomicPosition pos, GenomicPosition oldseqlen, const DnaSequence& newseq, char dir=0) : _direction(dir), _region(pos, oldseqlen), _newSeq(newseq) {}
	bool operator< (const Variant& other) const { return  _region.position() < other.position(); }
	GenomicPosition length() const { return max(_region.length(), (GenomicPosition) _newSeq.length()); }
	void clear() { _region.clear(); _newSeq.clear(); _text.clear(); }
	Variant operator+(GenomicPosition posOffset) const { Variant ret(*this); ret.position() += posOffset; return ret; }
	Variant operator-(GenomicPosition posOffset) const { Variant ret(*this); ret.position() -= posOffset; return ret; }
	bool operator==(const Variant& other) const {
		return _region == other._region and _newSeq == other._newSeq;
	}

	GenomicPosition position() const { return _region.position(); }
	GenomicPosition& position() { return _region.position(); }
	GenomicPosition oldSeqLength() const { return _region.length(); }
	GenomicPosition& oldSeqLength() { return _region.length(); }
	GenomicPosition newSeqLength() const { return _newSeq.length(); }
	const DnaSequence& newSeq() const { return _newSeq; }
	DnaSequence& newSeq() { return _newSeq; }
	const String& text() const { return _text; }
	String& text() { return _text; }
	const GenomicRegion& region() const { return _region; }

	void setNewSeq(const DnaSequence& ns) { _newSeq = ns; }
	void setText(const String& t) { _text = t; }
	bool isSNP() const { return _region.length() == 1 and _newSeq.length() == 1; }

	static DnaSequence applyVariants(const DnaSequence& original, const VariantSet& variants, /* subtracted from variant pos: */ GenomicPosition posOffset, char direction = 0);

	// find variants between two known sequences, max length 65535 bp
	static bool findVariants(const DnaSequence &a, const DnaSequence &b, VariantSet& variants, /* added to all variants */ GenomicPosition posOffset = 0);

#if 0
	// find variants for longer sequences, where sequence b is part of sequenceIndex'th sequence of genomeBfragments
	static void findVariants(const DnaSequence& a, const DnaSequence& b,
		const GenomeFragmentArray& genomeBfragments, 
		const ChromosomeIndex sequenceIndex, /* returned */ list<Variant> &variants);
#endif
	// returns alignment score for two equal sequences of given length:
	static int32_t perfectMatchAlignmentScore(uint16_t length);

	// calculates the alignment score given a list of variants (ie the alignment itself has already been done)
	// since it can't check positions of variants, it assumes all variants in the list are in the sequences concerned
	static int32_t alignmentScore(uint16_t lengthA, uint16_t lengthB, const list<Variant>&);

	// outputs alignment of two sequences up to 32767 bp long. Variants must be sorted
	static void showAlignment(ostream& o, const DnaSequence& a, const DnaSequence& b, const list<Variant>& variants, GenomicPosition posOffset = 0, uint16_t lineLength = 0);
	static void showAlignment(ostream& o, const DnaSequence& a, const DnaSequence& b, uint16_t lineLength = 0) {
		list<Variant> v;
		findVariants(a, b, v);
		showAlignment(o, a, b, v, 0, lineLength);
	}
	static bool parseAlignment(const String& aa, const String& bb, /* returned */ list<Variant>& variants);

	// find number of bases different:
	static uint32_t numberOfDifferences(const DnaSequence &seq1, const DnaSequence &seq2);
	static uint32_t numberOfDifferences(const list<Variant>&);

	/* loads variants from in stream and returns header and list of variants */
	static String loadVariants(const Genome&, istream&, /* output: */ VariantSet&);

protected:
	// Smith Waterman alignment
	static bool findVariantsSW(const char* const aa, const uint16_t al,
		const char* const bb, const uint16_t bl,
		/* returned: */ list<Variant>& variants, const GenomicPosition posOffset);
	static void showAlignmentC(ostream& o, const char* aa, uint16_t al,
			const char* bb, uint16_t bl,
			const list<Variant>& variants, const GenomicPosition posOffset = 0, const uint16_t lineLength = 0);
};


class ContigId
{
public:
	char _direction;	// direction of assembly, 'u' or 'd'
	uint32_t _contigNumber;
	uint16_t _versionNumber;
	uint32_t _averageScore;
	GenomicPosition _contigStartPosition;
	GenomicPosition _contigLength;
	uint16_t _coverageBefore;
	uint16_t _coverageAfter;
	ContigId(char direction, uint32_t contigNumber, uint32_t versionNumber, 
			uint32_t averageScore, GenomicPosition contigStart, GenomicPosition contigLength,
			uint16_t cb, uint16_t ca) :
		_direction(direction), _contigNumber(contigNumber), _versionNumber(versionNumber),
		_averageScore(averageScore), _contigStartPosition(contigStart), _contigLength(contigLength),
		_coverageBefore(cb), _coverageAfter(ca) {}
};


class PrinsesVariant : public Variant
{
public:
	list<ContigId> _contigIds;

	PrinsesVariant(GenomicPosition pos, GenomicPosition oldseqlen, const DnaSequence& newseq, char direction, uint32_t contigNumber, uint32_t versionNumber, 
			uint32_t averageScore, GenomicPosition contigStart, GenomicPosition contigLength,
			uint16_t coverageBefore, uint16_t coverageAfter)
		: Variant(pos, oldseqlen, newseq)
		{ _contigIds.push_back(ContigId(direction, contigNumber, versionNumber, averageScore, contigStart, contigLength, coverageBefore, coverageAfter)); }
	PrinsesVariant(GenomicPosition pos, GenomicPosition oldseqlen, const DnaSequence& newseq, const PrinsesVariant& templ)
		: Variant(pos, oldseqlen, newseq), _contigIds(templ._contigIds) {}

	/* loads variants from in stream and returns header and list of variants */
	static String loadVariants(const Genome& genome, istream&, /* output: */ map<String, list<PrinsesVariant> >&);
};


inline ostream& operator<<(ostream& o, const Variant& s)
{
	o << s.position() + 1
		<< '\t' << s.oldSeqLength()
		<< '\t' << s.newSeq()
		<< endl << s.text();
	return o;
}

inline ostream& operator<<(ostream& o, const PrinsesVariant& s)
{
	o	<< s.position() + 1
		<< '\t' << s.oldSeqLength()
		<< '\t' << s.newSeq()
		<< '\t';
	list<ContigId>::const_iterator i(s._contigIds.begin());
	while (true) {
		o << i->_direction << '.' << i->_contigNumber << '.' << i->_versionNumber
		<< '.' << i->_averageScore
		<< '.' << i->_contigStartPosition
		<< '.' << i->_contigLength
		<< '.' << i->_coverageBefore
		<< '.' << i->_coverageAfter;
		i++;
		if (i == s._contigIds.end()) break;
		o << ", ";
		
	}
	o	<< '\n' << s.text();
	return o;
}

#endif
