/*

Copyright 2010 Andreas Massouras

This file is part of Primer-Initiated Sequence Synthesis (PrInSeS)

For license information, please refer to the LICENSE document included in the distribution.
By downloading or using this software, you accept and are bound by the terms of the license.

*/

#ifndef VECTOR_H
#define VECTOR_H

#include "Util.h"
#include <math.h>

//#define DEBUG_SORT
#define SORT_METHOD_1

class VectorI;
class VectorD;

template <class T>
class Vector
{
public:
	Vector(uint32_t dimension = 0) : _dimension(dimension), _array(NULL) {
		if (dimension) _array = new T [dimension];
	}
	Vector(uint32_t dimension, const T& initialValue) : _dimension(dimension) {
		_array = new T [dimension];
		set(initialValue);
	}
	Vector(const Vector& other) : _dimension(other._dimension) {
		_array = new T [_dimension];
		for (uint32_t d = 0; d < _dimension; ++d) _array[d] = other._array[d];
	}
	const Vector& operator=(const Vector& other) {
		if (_dimension != other._dimension) resize(other._dimension);
		for (uint32_t d = 0; d < _dimension; ++d) _array[d] = other._array[d];
		return *this;
	}
	const Vector& operator=(const T& value) { set(value); return *this; }

	void resize(uint32_t newDimension = 0) {
		if (newDimension == _dimension) return;

		if (_array) delete [] _array;
		_dimension = newDimension;
		if (newDimension) {
			_array = new T [newDimension];
		} else {
			_array = NULL;
		}
	}
	void freeAll() {	// for Vector of pointers
		if (!_dimension) return;
		for (uint32_t d = 0; d < _dimension; ++d) delete _array[d];
		delete [] _array;
		_array = NULL;
	}
	~Vector() { resize(0); }

	void set(const T& value) {
		for (uint32_t d = 0; d < _dimension; ++d) _array[d] = value;
	}

	T& element(uint32_t d) {
#ifdef DEBUG
		if (d >= _dimension) {
			cerr << "Element " << d << " out of bounds for vector of dimension " << _dimension << endl;
			abort();
		}
#endif
		return _array[d];
	}
	const T& element(uint32_t d) const {
#ifdef DEBUG
		if (d >= _dimension) {
			cerr << "Element " << d << " out of bounds for vector of dimension " << _dimension << endl;
			abort();
		}
#endif
		return _array[d];
	}
	T& operator[](uint32_t d) {
#ifdef DEBUG
		if (d >= _dimension) {
			cerr << "Element " << d << " out of bounds for vector of dimension " << _dimension << endl;
			abort();
		}
#endif
		return _array[d];
	}
	const T& operator[](uint32_t d) const {
#ifdef DEBUG
		if (d >= _dimension) {
			cerr << "Element " << d << " out of bounds for vector of dimension " << _dimension << endl;
			abort();
		}
#endif
		return _array[d];
	}
	uint32_t dimension() const { return _dimension; }
	uint32_t size() const { return _dimension; }
	T sum(const T& zero = 0) const { T ret(zero); for (uint32_t d = 0; d < _dimension; ++d) ret += _array[d]; return ret; }
	T max() const {
		assert(_dimension);
		T m(_array[0]);
		for (uint32_t d = 1; d < _dimension; ++d)
			if (m < _array[d]) m = _array[d];
		return m;
	}
	T max(const T& seed) const {
		T m(seed);
		for (uint32_t d = 0; d < _dimension; ++d)
			if (m < _array[d]) m = _array[d];
		return m;
	}
	T min() const {
		assert(_dimension);
		T m(_array[0]);
		for (uint32_t d = 1; d < _dimension; ++d)
			if (_array[d] < m) m = _array[d];
		return m;
	}
	T min(const T& seed) const {
		T m(seed);
		for (uint32_t d = 0; d < _dimension; ++d)
			if (_array[d] < m) m = _array[d];
		return m;
	}

	const Vector& operator+=(const Vector& other) {
		assert(_dimension == other._dimension);
		for (uint32_t d = 0; d < _dimension; ++d) _array[d] += other._array[d];
		return *this;
	}
	const Vector& operator-=(const Vector& other) {
		assert(_dimension == other._dimension);
		for (uint32_t d = 0; d < _dimension; ++d) _array[d] -= other._array[d];
		return *this;
	}
	const Vector& operator+=(const T& value) {
		for (uint32_t d = 0; d < _dimension; ++d) _array[d] += value;
		return *this;
	}
	const Vector& operator-=(const T& value) {
		for (uint32_t d = 0; d < _dimension; ++d) _array[d] -= value;
		return *this;
	}
	const Vector& operator*=(const T& value) {
		for (uint32_t d = 0; d < _dimension; ++d) _array[d] *= value;
		return *this;
	}
	T dotProduct(const Vector& other, const T& zero = 0) const {
		assert(_dimension == other._dimension);
		T ret(zero);
		for (uint32_t d = 0; d < _dimension; ++d) ret += _array[d] * other._array[d];
		return ret;
	}
	void swap(uint32_t i, uint32_t j) {
		T tmp(_array[i]);
		_array[i] = _array[j];
		_array[j] = tmp;
	}

	class LowHigh
	{
	public:
		uint32_t _low;
		uint32_t _high;
		LowHigh(uint32_t l, uint32_t h) : _low(l), _high(h) {}
		LowHigh() {}
		uint32_t middle() const { return (_low + _high) / 2; }
	};

	// sorts and returns a vector r where this->[r[i]] is sorted for 0 <= i < dimension
	// uses quicksort
	VectorI order(bool (*lessThan)(const T&, const T&) = NULL) const;

	// returns a vector r where each element is the rank of *this.
	// for each i, j (0 <= i, j < _dimension) if r[i] < r[j] then this->[i] < this->[j]
	// rank starts with 1
	VectorI rank() const;

	// same as rank() but averages for equal values
	VectorD averageRank() const;

	double mean() const {
		assert(_dimension);
		double m = 0.0;
		for (uint32_t i = 0; i < _dimension; ++i) {
			m += (double) _array[i];
		}
		return m / _dimension;
	}

	void meanAndVariace(/* returned */ double& mean_, double &variance_) const {
		const uint32_t n = _dimension;
		assert(n > 1);
		double sumsq = 0.0;
		mean_ = mean();
		for (uint32_t i = 0; i < n; ++i) {
			const double v = (double) _array[i] - mean_;
			sumsq += v * v;
		}
		variance_ = sumsq / (double) (n - 1);
	}

	double variance() const {
		double m, var;
		meanAndVariace(m, var);
		return var;
	}

	double standardDeviation() const { return sqrt(variance()); }

private:

	uint32_t _dimension;
	T *_array;
};


class VectorI
{
public:
	VectorI(uint32_t dimension = 0) : _dimension(dimension), _array(NULL) {
		if (dimension) _array = new int32_t [dimension];
	}
	VectorI(uint32_t dimension, int32_t initialValue) : _dimension(dimension), _array(NULL) {
		if (dimension) {
			_array = new int32_t [dimension];
			set(initialValue);
		}
	}
	VectorI(const VectorI& other) : _dimension(other._dimension), _array(NULL) {
//		cdbg << "copy" << endl;
		if (!_dimension) return;
		_array = new int32_t [_dimension];
		for (uint32_t d = 0; d < _dimension; ++d) _array[d] = other._array[d];
	}
	VectorI(int32_t *array, uint32_t dimension) : _dimension(dimension), _array(NULL) {
		if (dimension) {
			_array = new int32_t [_dimension];
			for (uint32_t d = 0; d < _dimension; ++d)
				_array[d] = array[d];
		}
	}
	const VectorI& operator=(const VectorI& other) {
//		cdbg << "copy=" << endl;
		if (_dimension != other._dimension) resize(other._dimension);
		for (uint32_t d = 0; d < _dimension; ++d) _array[d] = other._array[d];
		return *this;
	}
	const VectorI& operator=(int32_t value) { set(value); return *this; }

	void resize(uint32_t newDimension = 0) {
		if (_dimension == newDimension) return;
		if (_array) delete [] _array;
		_dimension = newDimension;
		if (newDimension > 0) {
			_array = new int32_t [newDimension];
		} else {
			_array = NULL;
		}
	}
	~VectorI() { resize(0); }

	void set(int32_t value) {
		for (uint32_t d = 0; d < _dimension; ++d) _array[d] = value;
	}

	int32_t& element(uint32_t d) {
#ifdef DEBUG
		if (d >= _dimension) {
			cerr << "Element " << d << " out of bounds for vector of dimension " << _dimension << endl;
			abort();
		}
#endif
		return _array[d];
	}
	int32_t element(uint32_t d) const {
#ifdef DEBUG
		if (d >= _dimension) {
			cerr << "Element " << d << " out of bounds for vector of dimension " << _dimension << endl;
			abort();
		}
#endif
		return _array[d];
	}
	int32_t& operator[](uint32_t d) {
#ifdef DEBUG
		if (d >= _dimension) {
			cerr << "Element " << d << " out of bounds for vector of dimension " << _dimension << endl;
			abort();
		}
#endif
		return _array[d];
	}
	int32_t operator[](uint32_t d) const {
#ifdef DEBUG
		if (d >= _dimension) {
			cerr << "Element " << d << " out of bounds for vector of dimension " << _dimension << endl;
			abort();
		}
#endif
		return _array[d];
	}
	uint32_t dimension() const { return _dimension; }
	uint32_t size() const { return _dimension; }
	int32_t sum() const { int32_t ret = 0; for (uint32_t d = 0; d < _dimension; ++d) ret += _array[d]; return ret; }
	int32_t max() const {
		assert(_dimension);
		int32_t m(_array[0]);
		for (uint32_t d = 1; d < _dimension; ++d)
			if (m < _array[d]) m = _array[d];
		return m;
	}
	int32_t min() const {
		assert(_dimension);
		int32_t m(_array[0]);
		for (uint32_t d = 1; d < _dimension; ++d)
			if (_array[d] < m) m = _array[d];
		return m;
	}
	const VectorI& operator+=(const VectorI& other) {
		assert(_dimension == other._dimension);
		for (uint32_t d = 0; d < _dimension; ++d) _array[d] += other._array[d];
		return *this;
	}
	const VectorI& operator-=(const VectorI& other) {
		assert(_dimension == other._dimension);
		for (uint32_t d = 0; d < _dimension; ++d) _array[d] -= other._array[d];
		return *this;
	}
	const VectorI& operator+=(int32_t value) {
		for (uint32_t d = 0; d < _dimension; ++d) _array[d] += value;
		return *this;
	}
	const VectorI& operator-=(int32_t value) {
		for (uint32_t d = 0; d < _dimension; ++d) _array[d] -= value;
		return *this;
	}
	const VectorI& operator*=(int32_t value) {
		for (uint32_t d = 0; d < _dimension; ++d) _array[d] *= value;
		return *this;
	}
	const VectorI& operator/=(int32_t value) {
		for (uint32_t d = 0; d < _dimension; ++d) _array[d] /= value;
		return *this;
	}
	int32_t dotProduct(const VectorI& other) const {
		assert(_dimension == other._dimension);
		int32_t ret = 0;
		for (uint32_t d = 0; d < _dimension; ++d) ret += _array[d] * other._array[d];
		return ret;
	}
	void swap(uint32_t i, uint32_t j) {
		int32_t tmp(_array[i]);
		_array[i] = _array[j];
		_array[j] = tmp;
	}

	// sorts and returns a vector r where this->[r[i]] is sorted for 0 <= i < dimension
	// uses quicksort
	VectorI order() const;

	// returns a vector r where each element is the rank of *this.
	// for each i, j (0 <= i, j < _dimension) if r[i] < r[j] then this->[i] < this->[j]
	// rank starts with 1
	VectorI rank() const {
		VectorI o = order();
		VectorI r(_dimension);
		for (uint32_t i = 0; i < _dimension; ++i) {
			r[o[i]] = i + 1;
		}
		return r;
	}

	VectorI sorted() const {
		VectorI ret(_dimension);
		VectorI o = order();
		for (uint32_t i = 0; i < _dimension; ++i)
			ret[i] = element(o[i]);
		return ret;
	}
	void sort();

	bool isSorted() const {
		for (uint32_t i = 1; i < _dimension; ++i) {
			if (_array[i] < _array[i-1]) return false;
		}
		return true;
	}


	// assuming vector is sorted, findSorted will return i where vec[i-1] <= d < vec[i]
	// 0 <= i <= dimension
	uint32_t findSorted(double d) const;

	// same as rank() but averages for equal values
	VectorD averageRank() const;

	double mean() const {
		assert(_dimension);
		double m = 0.0;
		for (uint32_t i = 0; i < _dimension; ++i) {
			m += (double) _array[i];
		}
		return m / _dimension;
	}

	void meanAndVariace(/* returned */ double& mean_, double &variance_) const;

	double variance() const {
		double m, var;
		meanAndVariace(m, var);
		return var;
	}

	double standardDeviation() const { return sqrt(variance()); }

	friend class VectorD;
private:

	uint32_t _dimension;
	int32_t* _array;
}; // VectorI


class VectorD
{
public:
	VectorD(uint32_t dimension = 0) : _dimension(dimension), _array(NULL) {
		if (dimension) _array = new double [dimension];
	}
	VectorD(uint32_t dimension, double initialValue) : _dimension(dimension), _array(NULL) {
		if (dimension) {
			_array = new double [dimension];
			set(initialValue);
		}
	}
	VectorD(const VectorD& other) : _dimension(other._dimension), _array(NULL) {
		if (!_dimension) return;
		_array = new double [_dimension];
		for (uint32_t d = 0; d < _dimension; ++d) _array[d] = other._array[d];
	}
	VectorD(const VectorI& other) : _dimension(other.dimension()), _array(NULL) {
		if (!_dimension) return;
		_array = new double [_dimension];
		for (uint32_t d = 0; d < _dimension; ++d) _array[d] = (double) other[d];
	}
	const VectorD& operator=(const VectorD& other) {
		if (_dimension != other._dimension) resize(other._dimension);
		for (uint32_t d = 0; d < _dimension; ++d) _array[d] = other._array[d];
		return *this;
	}
	const VectorD& operator=(const VectorI& other) {
		if (_dimension != other.dimension()) resize(other.dimension());
		for (uint32_t d = 0; d < _dimension; ++d) _array[d] = (double) other[d];
		return *this;
	}
	const VectorD& operator=(double value) { set(value); return *this; }

	void resize(uint32_t newDimension = 0) {
		if (_dimension == newDimension) return;
		if (_array) delete [] _array;
		_dimension = newDimension;
		if (newDimension > 0) {
			_array = new double [newDimension];
		} else {
			_array = NULL;
		}
	}
	~VectorD() { resize(0); }

	void set(double value) {
		for (uint32_t d = 0; d < _dimension; ++d) _array[d] = value;
	}

	double& element(uint32_t d) {
#ifdef DEBUG
		if (d >= _dimension) {
			cerr << "Element " << d << " out of bounds for vector of dimension " << _dimension << endl;
			abort();
		}
#endif
		return _array[d];
	}
	double element(uint32_t d) const {
#ifdef DEBUG
		if (d >= _dimension) {
			cerr << "Element " << d << " out of bounds for vector of dimension " << _dimension << endl;
			abort();
		}
#endif
		return _array[d];
	}
	double& operator[](uint32_t d) {
#ifdef DEBUG
		if (d >= _dimension) {
			cerr << "Element " << d << " out of bounds for vector of dimension " << _dimension << endl;
			abort();
		}
#endif
		return _array[d];
	}
	double operator[](uint32_t d) const {
#ifdef DEBUG
		if (d >= _dimension) {
			cerr << "Element " << d << " out of bounds for vector of dimension " << _dimension << endl;
			abort();
		}
#endif
		return _array[d];
	}
	uint32_t dimension() const { return _dimension; }
	uint32_t size() const { return _dimension; }
	double sum() const { double ret(0.0); for (uint32_t d = 0; d < _dimension; ++d) ret += _array[d]; return ret; }
	double max() const {
		assert(_dimension);
		double m(_array[0]);
		for (uint32_t d = 1; d < _dimension; ++d)
			if (m < _array[d]) m = _array[d];
		return m;
	}
	double min() const {
		assert(_dimension);
		double m(_array[0]);
		for (uint32_t d = 1; d < _dimension; ++d)
			if (_array[d] < m) m = _array[d];
		return m;
	}

	const VectorD& operator+=(const VectorD& other) {
		assert(_dimension == other._dimension);
		for (uint32_t d = 0; d < _dimension; ++d) _array[d] += other._array[d];
		return *this;
	}
	const VectorD& operator-=(const VectorD& other) {
		assert(_dimension == other._dimension);
		for (uint32_t d = 0; d < _dimension; ++d) _array[d] -= other._array[d];
		return *this;
	}
	const VectorD& operator+=(double value) {
		for (uint32_t d = 0; d < _dimension; ++d) _array[d] += value;
		return *this;
	}
	const VectorD& operator-=(double value) {
		for (uint32_t d = 0; d < _dimension; ++d) _array[d] -= value;
		return *this;
	}
	const VectorD& operator*=(double value) {
		for (uint32_t d = 0; d < _dimension; ++d) _array[d] *= value;
		return *this;
	}
	const VectorD& operator/=(double value) {
		for (uint32_t d = 0; d < _dimension; ++d) _array[d] /= value;
		return *this;
	}
	VectorD operator+(double value) const {
		VectorD ret(_dimension);
		for (uint32_t d = 0; d < _dimension; ++d)
			ret._array[d] = _array[d] + value;
		return ret;
	}
	VectorD operator-(double value) const {
		VectorD ret(_dimension);
		for (uint32_t d = 0; d < _dimension; ++d)
			ret._array[d] = _array[d] - value;
		return ret;
	}
	VectorD operator*(double value) const {
		VectorD ret(_dimension);
		for (uint32_t d = 0; d < _dimension; ++d)
			ret._array[d] = _array[d] * value;
		return ret;
	}
	VectorD operator/(double value) const {
		VectorD ret(_dimension);
		for (uint32_t d = 0; d < _dimension; ++d)
			ret._array[d] = _array[d] / value;
		return ret;
	}
	VectorD operator+(const VectorD& o) const {
		assert(_dimension == o._dimension);
		VectorD ret(_dimension);
		for (uint32_t d = 0; d < _dimension; ++d)
			ret._array[d] = _array[d] + o._array[d];
		return ret;
	}
	VectorD operator-(const VectorD& o) const {
		assert(_dimension == o._dimension);
		VectorD ret(_dimension);
		for (uint32_t d = 0; d < _dimension; ++d)
			ret._array[d] = _array[d] - o._array[d];
		return ret;
	}
	double dotProduct(const VectorD& other) const {
		assert(_dimension == other._dimension);
		double ret(0.0);
		for (uint32_t d = 0; d < _dimension; ++d) ret += _array[d] * other._array[d];
		return ret;
	}
	void swap(uint32_t i, uint32_t j) {
		double tmp(_array[i]);
		_array[i] = _array[j];
		_array[j] = tmp;
	}

	// sorts and returns a vector r where this->[r[i]] is sorted for 0 <= i < dimension
	// uses quicksort
	VectorI orderQSort() const;
	VectorI orderMSort() const;
	VectorI order() const { return _dimension > 10000 ? orderMSort() : orderQSort(); }
//	VectorI order() const { return orderQSort(); }

	// returns a vector r where each element is the rank of *this.
	// for each i, j (0 <= i, j < _dimension) if r[i] < r[j] then this->[i] < this->[j]
	// rank starts with 1
	VectorI rank() const;

	void sort();

	VectorD sorted() const {
/*
		VectorD ret(_dimension);
		VectorI o = order();
		for (uint32_t i = 0; i < _dimension; ++i)
			ret[i] = _array[o[i]];
		return ret;
*/
		VectorD ret(*this);
		ret.sort();
		return ret;
	}

	bool isSorted() const {
		for (uint32_t i = 1; i < _dimension; ++i) {
			if (_array[i] < _array[i-1]) return false;
		}
		return true;
	}

	// assuming vector is sorted, findSorted will return i where vec[i-1] <= d < vec[i]
	// 0 <= i <= dimension
	uint32_t findSorted(double d) const;

	// same as rank() but averages for equal values
	VectorD averageRank() const;

	double mean() const {
		assert(_dimension);
		double m = 0.0;
		for (uint32_t i = 0; i < _dimension; ++i) {
			m += _array[i];
		}
		return m / _dimension;
	}

	void meanAndVariace(/* returned */ double& mean_, double &variance_) const;

	double variance() const {
		double m, var;
		meanAndVariace(m, var);
		return var;
	}

	double standardDeviation() const { return sqrt(variance()); }

private:
	void msort(int32_t* const ret, int32_t* const tmp, const uint32_t n
#ifdef DEBUG_SORT
		, uint32_t& numComp
#endif
	) const;

	uint32_t _dimension;
	double *_array;
}; // VectorD


// sorts and returns a vector r where this->[r[i]] is sorted for 0 <= i < dimension
// uses quicksort
template <class T> VectorI
Vector<T>::order(bool (*lessThan)(const T&, const T&)) const
{
	if (_dimension <= 1) return VectorI(_dimension, 0);

	VectorI r(_dimension);
	for (uint32_t i = 0; i < _dimension; ++i) r[i] = i;

	Stack<LowHigh> stack;
	stack.push(LowHigh(0, _dimension-1));

	while (!stack.isEmpty()) {
		LowHigh lh = stack.pop();

		// trivial if only a few elements:
		if (lh._low + 4 >= lh._high) {
			while (lh._low < lh._high) {
				uint32_t min = lh._low;
				if (lessThan) {
					for (uint32_t i = lh._low + 1; i <= lh._high; ++i) {
						if (lessThan(_array[r[i]], _array[r[min]])) min = i;
					}
				} else {
					for (uint32_t i = lh._low + 1; i <= lh._high; ++i) {
						if (_array[r[i]] < _array[r[min]]) min = i;
					}
				}
				if (min != lh._low) r.swap(min, lh._low);
				lh._low++;
			}
			continue;
		}

		// partition:
		// first put the median of (low, mid, high) at the middle as a pivot:
		const uint32_t mid = lh.middle();
		if (lessThan) {
			if (lessThan(_array[r[mid]], _array[r[lh._low]])) {
				r.swap(mid, lh._low);
			}
			if (lessThan(_array[r[lh._high]], _array[r[mid]])) {
				r.swap(mid, lh._high);
				if (lessThan(_array[r[mid]], _array[r[lh._low]])) {
					r.swap(mid, lh._low);
				}
			}
		} else {
			if (_array[r[mid]] < _array[r[lh._low]]) {
				r.swap(mid, lh._low);
			}
			if (_array[r[lh._high]] < _array[r[mid]]) {
				r.swap(mid, lh._high);
				if (_array[r[mid]] < _array[r[lh._low]]) {
					r.swap(mid, lh._low);
				}
			}
		}

		const uint32_t pivotIndex(r[mid]);
		// at this stage _array[r[lh._low]] < _array[pivotIndex] and _array[r[lh._high]] > _array[pivotIndex]
		r[mid] = r[lh._high-1];
#ifdef SORT_METHOD_1		// "collapsing walls" method:
		uint32_t low = lh._low + 1;
		uint32_t high = lh._high - 2;
		uint32_t pivot;
		if (lessThan) {
			while (low < high) {
				while (low < high and lessThan(_array[r[low]], _array[pivotIndex]))
					low++;
				while (low < high and lessThan(_array[pivotIndex], _array[r[high]]))
					high--;

				if (low < high) {
					r.swap(low, high);
					low++;
					high--;
				}
			}
			pivot = lessThan(_array[r[low]], _array[pivotIndex]) ? low + 1 : low;
		} else {
			while (low < high) {
				while (low < high and _array[r[low]] < _array[pivotIndex])
					low++;
				while (low < high and _array[pivotIndex] < _array[r[high]])
					high--;

				if (low < high) {
					r.swap(low, high);
					low++;
					high--;
				}
			}
			pivot = _array[r[low]] < _array[pivotIndex] ? low + 1 : low;
		}

#else	// method 2 - this can be disastrously slow in some cases
		uint32_t pivot = lh._low + 1;
		if (lessThan) {
			for (uint32_t i = lh._low + 1; i < lh._high - 1; ++i) {
				if (lessThan(_array[r[i]], _array[pivotIndex])) {
					if (i != pivot) r.swap(i, pivot);
					pivot++;
				}
			}
		} else {
			for (uint32_t i = lh._low + 1; i < lh._high - 1; ++i) {
				if (_array[r[i]] < _array[pivotIndex]) {
					if (i != pivot) r.swap(i, pivot);
					pivot++;
				}
			}
		}
#endif // SORT_METHOD_1
		if (pivot > lh._low + 1)
			stack.push(LowHigh(lh._low, pivot-1));

		if (pivot + 1 < lh._high) {
			r[lh._high-1] = r[pivot];
			stack.push(LowHigh(pivot+1, lh._high));
		}
		r[pivot] = pivotIndex;
	}
	return r;
}


// returns a vector r where each element is the rank of *this.
// for each i, j (0 <= i, j < _dimension) if r[i] < r[j] then this->[i] < this->[j]
// rank starts with 1
template <class T>
VectorI
Vector<T>::rank() const
{
	VectorI o = order();
	VectorI r(_dimension);
	for (uint32_t i = 0; i < _dimension; ++i) {
		r[o[i]] = i + 1;
	}
	return r;
}


// same as rank() but averages for equal values
template <class T>
VectorD
Vector<T>::averageRank() const
{
	VectorI o = order();
	const uint32_t n = _dimension;
	VectorD ar(n);
	for (uint32_t i = 0; i < n; i++) {
		uint32_t ord = o[i];
		uint32_t sumRank = i + 1;
		uint32_t num = 1;
		uint32_t j = i;
		while (j) {
			j--;
			if (_array[o[j]] == _array[ord]) {
				sumRank += j + 1;
				num++;
			}
		}
		const double averageRank = ((double) sumRank) / num;
		for (j = i; num; j--, num--) {
			ar[o[j]] = averageRank;
		}
	}
	return ar;
}

#endif
