/*

Copyright 2011 Andreas Massouras

This file is part of Primer-Initiated Sequence Synthesis (PrInSeS)

For license information, please refer to the LICENSE document included in the distribution.
By downloading or using this software, you accept and are bound by the terms of the license.

*/

#include <iostream>
#include <fstream>
#include <stdlib.h>
#include "Util.h"
#include "DnaSequence.h"
#include "Variant.h"

using namespace std;

namespace PrinsesCoverage { extern void main(int argc, char* argv[]); }
namespace PrinsesG { extern void main(int argc, char* argv[]); }
namespace PrinsesPrepare { extern void main(int argc, char* argv[]); }

namespace PrinsesApplyVariants
{

static String programName;

// Command line parameters:
static bool printVersion = false;

void
usage()
{
	cerr << "Usage: " << programName << " genome.fa variants.txt [u|d] > new.sequences.fa" << endl;
	cerr << "Applies variants in variants.txt to genome.fa" << endl;
	exit(1);
}


void
main(int argc, char* argv[])
{
	stringstream signature;
	signature << argv[0] << ' ' << argv[1];
	programName = signature.str();
	argc--; argv++;
	for (int i = 1; i < argc; ++i)
		signature << ' ' << argv[i];
#ifdef DEBUG
	signature << " debug";
#endif

	time_t time0 = time(NULL);
	cerr << Util::getLocalTime(time0) << " " << signature.str() << endl;

	// Get command line arguments:
	map<String, bool*> flags;
	map<String, int*> intOptions;
	map<String, String*> stringOptions;
	flags["version"] = &printVersion;
	char direction = 0;
	vector<String> argV = Util::getCommandLineOptions(flags, intOptions, stringOptions, argc, argv, usage);
	if (printVersion) {
		cerr << "built on " << __DATE__ << " " << __TIME__ << endl;
		exit(0);
	}
	if (argV.size() != 2) {
		if (argV.size() == 3 and (argV[2][0] == 'u' or argV[2][0] == 'd')) {
			direction = argV[2][0];
		} else usage();
	}

	cerr << Util::getLocalTime(time(NULL)) << " loading genome...\n";
	Genome genome(argV[0].c_str());
	ifstream varFile(argV[1].c_str());
	if (varFile.fail()) {
		cerr << "Can't open file " << argV[2];
		exit(1);
	}
	VariantSet variants;
	String header = Variant::loadVariants(genome, varFile, variants);

	cerr << Util::getLocalTime(time(NULL)) << " applying " << variants.size() << " variants to genome..." << endl;
	genome.applyVariants(variants, direction);
	cerr << Util::getLocalTime(time(NULL)) << " writing genome..." << endl;
	cout << genome;

	// Wind down:
	time_t time1 = time(NULL);
	cerr << Util::getLocalTime(time1) << " " << programName << " done in " << Util::hms(time1-time0) << endl;
}


}; // namespace PrinsesApplyVariants


namespace PrinsesWrapper
{

class App
{
public:
	string _name;
	void (* _function)(int argc, char* argv[]);
	App(const char* n, void (* f)(int argc, char* argv[]))
		: _name(n), _function(f) {}
	App() : _function(NULL) {}
};

static const App apps[] = {
	App("prepare", PrinsesPrepare::main), 
	App("assemble", PrinsesG::main), 
	App("validate", PrinsesCoverage::main), 
	App("apply", PrinsesApplyVariants::main), 
	App()
};


static void
usage(const char* progname)
{
	cerr << "Usage:" << endl;
	for (const App* u = apps; u->_name.length(); ++u) {
		 cerr << '\t' << progname << ' ' << u->_name << endl;
	}
	cerr << "Run each option without arguments for more information" << endl;
	exit(1);
}


static void
main(int argc, char* argv[])
{
	if (argc < 2) usage(argv[0]);
	for (const App* u = apps; u->_name.length(); ++u) {
		if (u->_name == argv[1]) {
			u->_function(argc, argv);
			return;
		}
	}
	usage(argv[0]);
}


}; // namespare PrinsesWrapper


int
main(int argc, char* argv[])
{
	PrinsesWrapper::main(argc, argv);
	return 0;
}
