/* 

Copyright 2010 Andreas Massouras

This file is part of Primer-Initiated Sequence Synthesis (PrInSeS)

For license information, please refer to the LICENSE document included in the distribution.
By downloading or using this software, you accept and are bound by the terms of the license.

*/


#include "Util.h"
#include "Genome.h"
#include "SamPileup.h"
#include "Variant.h"
#include <iostream>
#include <fstream>
#include <iomanip>
#include <assert.h>

using namespace std;

namespace PrinsesCoverage
{

static String programName;

// Command line parameters:
static bool printVersion = false;
static bool doSNPs = true;
static int minDepth = 3;

// Global constants:
static const GenomicPosition flankingLength = 5;

// Global variables:

static void
usage()
{
	cerr << "Usage: " << programName << " prinsesDirectory genome.fa alignments.bam [script]" << endl;
	cerr << "   creates file variants.verified.txt" << endl;
	cerr << "   runs 'script' with the PrInSeS output genome as reference" << endl;
	cerr << "   script should take a fasta file as argument and produce a sorted bam file (must build index if necessary)" << endl;
	cerr << "   example: 'script upstream.fa' should produce 'upstream.fa.bam'" << endl;
	cerr << "   if 'script' is not given, we assume the output bam file has already been created" << endl;
	cerr << endl << "Options:" << endl;
	cerr << "  -version       Prints version and exits" << endl;
	cerr << "  -nosnps        Ignore SNPs in final statistics" << endl;
	exit(1);
}


static void
execute(const char* command)
{
	cerr << Util::getLocalTime(time(NULL)) << " " << command << endl;
	if (system(command) == -1) {
		cerr << "system() returned error - aborting" << endl;
		exit(1);
	}
}


void
doOneChromosome(const ChromosomeIndex chromoIndex, VariantSet& variants, ostream &outFile,
	const SamPileup &referenceSam, const SamPileup& upstreamSam, const SamPileup& downstreamSam,
	double& sumBefore, double& sumAfter, uint32_t& numBetter, uint32_t& numWorse, uint32_t& numUnchd, uint32_t& variantCount)
{
	VariantSet::iterator vi;
	variants.sort();

	const Genome& referenceGenome = referenceSam.referenceGenome();
	const Genome& upstreamGenome = upstreamSam.referenceGenome();
	const Genome& downstreamGenome = downstreamSam.referenceGenome();
	const int64_t chromoStart = referenceGenome.chromoStart(chromoIndex);

	int64_t upOffset = (int64_t) upstreamGenome.chromoStart(chromoIndex) - chromoStart;
	int64_t downOffset = (int64_t) downstreamGenome.chromoStart(chromoIndex) - chromoStart;
//	cdbg << "qq chromoStart = " << chromoStart
//		<< " up = " << upstreamGenome.chromoStart(chromoIndex)
//		<< " down = " << downstreamGenome.chromoStart(chromoIndex) << endl;

	for (vi = variants.begin(); vi != variants.end(); ++vi) {
		const GenomicPosition pos = vi->position();
//		cdbg << "qq pos = " << pos << " = " << referenceGenome.toText(pos)
//			<< ", upOffset = " << upOffset << " = " << upstreamGenome.toText(pos + upOffset)
//			<< ", downOffet = " << downOffset  << " = " << downstreamGenome.toText(pos + downOffset)
//			<< endl;
		const GenomicPosition oldSeqLength = vi->oldSeqLength();
		const GenomicPosition newSeqLength = vi->newSeqLength();
		const char direction = vi->_direction;
		bool inStatistics = doSNPs or !vi->isSNP();
		const double beforeDepth = referenceSam.averageReadDepth(pos, oldSeqLength, flankingLength);
		const double beforeBP1 = ((double) (referenceSam.depthAtPosition(pos - 1) + referenceSam.depthAtPosition(pos))) / 2.0;
		const double beforeBP2 = oldSeqLength == 0 ? beforeBP1 : ((double) (referenceSam.depthAtPosition(pos + oldSeqLength - 1) + referenceSam.depthAtPosition(pos + oldSeqLength))) / 2.0;
		if (inStatistics) sumBefore += beforeDepth;
		double uDepth = 0.0, dDepth = 0.0, afterDepth = 0.0;
		if (direction == 'u' or direction == 'b') {
			const ChromosomeIndex ci2 = upstreamGenome.chromoIndex(pos + upOffset);
			if (ci2 != chromoIndex) {
				cerr << "Major Problem: upstream index " << ci2 << " different from " << chromoIndex << ", pos = " << pos << " = " << referenceGenome.toText(pos) << ", upOffset = " << upOffset << endl;
			}
			afterDepth = uDepth = upstreamSam.averageReadDepth(pos + upOffset, newSeqLength, flankingLength);
			const double uBP1 = ((double) (upstreamSam.depthAtPosition(pos + upOffset - 1) + upstreamSam.depthAtPosition(pos + upOffset))) / 2.0;
			const double uBP2 = newSeqLength == 0 ? uBP1 : ((double) (upstreamSam.depthAtPosition(pos + upOffset + newSeqLength - 1) + upstreamSam.depthAtPosition(pos + upOffset + newSeqLength))) / 2.0;
			stringstream out;
			out << "# read depth: before " << beforeDepth << ", after ("
				<< upstreamGenome.chromoPos(pos + upOffset) + 1 << ") "
				<< uDepth << " (upstream)" << endl;
			out << "# at first breakpoint: " << beforeBP1 << " -> " << uBP1 << ", at second breakpoint: " << beforeBP2 << " -> " << uBP2 << endl;
			vi->text() += out.str();

			upOffset += newSeqLength;
			upOffset -= oldSeqLength;
		}
		if (direction == 'd' or direction == 'b') {
			const ChromosomeIndex ci2 = downstreamGenome.chromoIndex(pos + downOffset);
			if (ci2 != chromoIndex) {
				cerr << "Major Problem: downstream index " << ci2 << " different from " << chromoIndex << ", pos = " << pos << " = " << referenceGenome.toText(pos) << ", downOffset = " << downOffset << endl;
			}
			assert(downstreamGenome.chromoIndex(pos + downOffset) == chromoIndex);
			afterDepth = dDepth = downstreamSam.averageReadDepth(pos + downOffset, newSeqLength, flankingLength);
			const double dBP1 = ((double) (downstreamSam.depthAtPosition(pos + downOffset - 1) + downstreamSam.depthAtPosition(pos + downOffset))) / 2.0;
			const double dBP2 = newSeqLength == 0 ? dBP1 : ((double) (downstreamSam.depthAtPosition(pos + downOffset + newSeqLength - 1) + downstreamSam.depthAtPosition(pos + downOffset + newSeqLength))) / 2.0;
			stringstream out;
			out << "# read depth: before " << beforeDepth << ", after ("
				<< downstreamGenome.chromoPos(pos + downOffset) + 1 << ") "
				<< dDepth << " (downstream)" << endl;
			out << "# at first breakpoint: " << beforeBP1 << " -> " << dBP1 << ", at second breakpoint: " << beforeBP2 << " -> " << dBP2 << endl;
			vi->text() += out.str();

			downOffset += newSeqLength;
			downOffset -= oldSeqLength;
		}
		if (direction == 'b') afterDepth = (uDepth + dDepth) / 2.0;
		if (inStatistics) {
			sumAfter += afterDepth;
			if (afterDepth - beforeDepth > 1.0) {
				numBetter++;
			} else if (afterDepth - beforeDepth < -1.0) {
				numWorse++;
			} else {
				numUnchd++;
			}
			variantCount++;
		}
		referenceGenome.output(outFile, *vi);
	}
}


void
main(int argc, char* argv[])
{
	Util::setThousandSeparator(cerr);
	cerr << fixed << setprecision(1);
	Util::setThousandSeparator(cout);

	stringstream signature;
	signature << argv[0] << ' ' << argv[1];
	programName = signature.str();
	argc--; argv++;
	for (int i = 1; i < argc; ++i)
		signature << ' ' << argv[i];
#ifdef DEBUG
	signature << " debug";
#endif

	time_t time0 = time(NULL);
	cerr << Util::getLocalTime(time0) << " " << signature.str() << endl;


	// Get command line arguments:
	map<String, bool*> flags;
	map<String, int*> intOptions;
	map<String, String*> stringOptions;
	flags["version"] = &printVersion;
	flags["snps"] = &doSNPs;
	vector<String> argV = Util::getCommandLineOptions(flags, intOptions, stringOptions, argc, argv, usage);
	if (printVersion) {
		cerr << "built on " << __DATE__ << " " << __TIME__ << endl;
		exit(0);
	}
	if (argV.size() < 3) usage();

	uint32_t argC = 0;
	String outputDirectory = argV[argC++];
	String genomeFile = argV[argC++];
	String referenceSamFilename = argV[argC++];
	String script;
	if (argC < argV.size()) {
		script = argV[argC++];
		if (script[0] != '/') {
			script = Util::getCurrentDirectory() + '/' + script;
		}
	}
	if (argC != argV.size()) usage();

	cerr << Util::getLocalTime(time(NULL)) << " loading genome\n";
	Genome referenceGenome(genomeFile.c_str());

	cerr << Util::getLocalTime(time(NULL)) << " loading variants\n";
	VariantSet variants;
	String header;
	{
		String varFilename = outputDirectory + "/variants.txt";
		ifstream varFile(varFilename.c_str());
		if (varFile.fail()) {
			cerr << "Can't open file " << varFilename << endl;
			exit(1);
		}
		header = Variant::loadVariants(referenceGenome, varFile, variants);
	}
	ofstream outFile((outputDirectory + "/variants.verified.txt").c_str());
	if (outFile.fail()) {
		cerr << programName << ": Can't open file " << outputDirectory << "/variants.verified.txt for writing"
#ifdef DEBUG
			<< " in " << __PRETTY_FUNCTION__ 
			<< " " << __FILE__ << " line " << __LINE__ 
#endif
			<< endl;
		exit(1);
	}
	outFile << fixed << setprecision(1) << "# " << Util::getLocalTime(time0) << ' ' << signature << endl << header;

	cerr << Util::getLocalTime(time(NULL)) << " loading prinses upstream and downstream sequences\n";
	Genome upstreamGenome((outputDirectory + "/upstream.fa").c_str());
	Genome downstreamGenome((outputDirectory + "/downstream.fa").c_str());
	if (!script.empty()) {
		stringstream upstreamCommand;
		upstreamCommand << script << ' ' << outputDirectory << "/upstream.fa";
		execute(upstreamCommand.str().c_str());
		stringstream downstreamCommand;
		downstreamCommand << script << ' ' << outputDirectory << "/downstream.fa";
		execute(downstreamCommand.str().c_str());
	}
//	cdbg << "REF:" << endl;
//	referenceGenome.outputStats(cdbg);
//	cdbg << "UP:" << endl;
//	upstreamGenome.outputStats(cdbg);
//	cdbg << "DOWN:" << endl;
//	downstreamGenome.outputStats(cdbg);
	SamPileup referenceSam(referenceSamFilename.c_str(), referenceGenome, minDepth);
	SamPileup upstreamSam((outputDirectory + "/upstream.fa.bam").c_str(), upstreamGenome, minDepth);
	SamPileup downstreamSam((outputDirectory + "/downstream.fa.bam").c_str(), downstreamGenome, minDepth);

	double sumBefore = 0.0;
	double sumAfter = 0.0;
	uint32_t numBetter = 0;
	uint32_t numWorse = 0;
	uint32_t numUnchd = 0;
	uint32_t variantCount = 0;

	VariantSet::iterator vi;
	const ChromosomeIndex numChromosomes = referenceGenome.numberChromosomes();
	VariantSet chromoVariants[numChromosomes];
	for (vi = variants.begin(); vi != variants.end(); ++vi) {
		const GenomicPosition pos = vi->position();
		const ChromosomeIndex ci = referenceGenome.chromoIndex(pos);
		chromoVariants[ci].push_back(*vi);
	}
	for (ChromosomeIndex ci = 0; ci < numChromosomes; ++ci) {
		doOneChromosome(ci, chromoVariants[ci], outFile, referenceSam, upstreamSam, downstreamSam,
			sumBefore, sumAfter, numBetter, numWorse, numUnchd, variantCount);
	}
	outFile.close();

	cerr	<< "Genome size:\t" << referenceGenome.length() << endl
		<< "Upstream genome:\t" << upstreamGenome.length() << endl
		<< "Downstream genome:\t" << downstreamGenome.length() << endl
		<< "Number of variants:\t" << variants.size() << endl
		<< "Number of variants compared:\t" << variantCount << endl;
	if (variantCount)
		cerr	<< "Number where read depth increased by at least 1:\t" << numBetter << endl
			<< "Number where read depth decreased by at least 1:\t" << numWorse << endl
			<< "Number where read depth changed by 1 or less:\t" << numUnchd << endl
			<< "Average read depth before accross variants:\t" << sumBefore / variantCount << endl
			<< "Average read depth after accross variants:\t" << sumAfter / variantCount << endl;

	// Wind down:
	time_t time1 = time(NULL);
	cerr << Util::getLocalTime(time1) << " " << programName << " done in " << Util::hms(time1-time0) << endl;
}

}; // namespace PrinsesCoverage
