/*

Copyright 2010 Andreas Massouras

This file is part of Primer-Initiated Sequence Synthesis (PrInSeS)

For license information, please refer to the LICENSE document included in the distribution.
By downloading or using this software, you accept and are bound by the terms of the license.



This program is NOT suitable for multi-threading.
Every effort has been made to reduce memory used for large read sets

*/

#include <fstream>
#include <iomanip>
#include <sys/stat.h>
#include <zlib.h>
#include "Reads.h"
#include "Util.h"
extern "C" {
#include "sam.h"
};

using namespace std;

namespace PrinsesPrepare
{

static String programName;

// Global variables:
static bool useScores = true;
static bool printVersion = false;
static int minLength = 20;
static int maxLength = 50000;
static int minAverageScore = 0;
static bool discardTooLong = true;
static int maxRepeats = 2;

// Global Constants:
static const unsigned lineLength = 32768;
static char readSequence[lineLength];	// this makes program not threadable, but can be moved

static const bool oldVersion = false;

static void
usage()
{
	if (oldVersion) {
		cerr << "Usage: " << programName << " [options] illumina|phred|fasta reads.file1 .... > reads.prinses\n";
		cerr << "illumina         fastq files with illumina 1.3+ quality scores\n";
		cerr << "phred            fastq files with phred quality scores\n";
		cerr << "fasta            fasta files\n";
	} else {
		cerr << "Usage: " << programName << " [options] alignments.bam > reads.prinses\n";
	}
	cerr << "\nOptions:\n";
	cerr << "  -version       Prints version and exits\n";
        cerr << "  -maxRepeats n  Discard reads that map to more than n loci (default " << maxRepeats << ")\n";
        cerr << "  -minLength n   Discard reads with length less than n (default " << minLength << ")\n";
        cerr << "  -maxLength n   Discard reads with length more than n (default " << maxLength << ")\n";
	cerr << "  -noscores      Don't include scores in the output file\n";
	cerr << "  -filter n      Discard reads whose average score is below n\n";

	exit(1);
}


typedef enum { Illumina = 0, Phred = 1, Fasta = 2, Raw = 3 } ReadAndScoreType;
const char qualityScoreBase[] = { 64, 33, 0, 4 };
const char qualityScoreScaling[] = { 2, 2, 1, 4 };


// This class has variable size, for member _score
class ScoreList
{
public:
	// number of bytes used for scores of a read of readLength bases, must be kept in line with data members
	static uint32_t byteSize(uint16_t readLength) { return (readLength+1)/2 + sizeof(ScoreList); }

	ScoreList(ReadAndScoreType, const char* score, uint16_t readLen, ScoreList* next = NULL);
	ScoreList* next() { return _next; }
	uint8_t score(uint16_t i) const  {
		return (i % 2 == 0) ? _score[i/2] / 16 : _score[i/2] & 15;
	}
	const uint8_t* rawScore() const  { return _score; }

	void *operator new(size_t s, uint32_t readLen) { return SimpleAllocator::allocate(byteSize(readLen)); }
	void operator delete(void *p) { SimpleAllocator::free((uint8_t*) p); }

	uint64_t memorySize(uint16_t readLen) const {
		const ScoreList* n = this;
		uint64_t ret = 0;
		uint32_t bs = byteSize(readLen);
		while (n) {
			ret += bs; 
			n = n->_next;
		}
		return ret;
	}

private:
	ScoreList* _next;
	uint8_t _score[];		// variable size
};


ScoreList::ScoreList(ReadAndScoreType scoreType, const char* score, uint16_t readLen, ScoreList* next)
{
	char scoreBase = qualityScoreBase[scoreType];
	char scoreScaling = qualityScoreScaling[scoreType];
	uint8_t* thisScore = _score;

	// convert score into a 4-bit number:
	uint16_t i;
	for (i = 0; i+1 < readLen; i += 2) {
		uint8_t c1, c2;
		if (score[i] > scoreBase) {
			c1 = (score[i] - scoreBase) / scoreScaling;
			if (c1 > 15) c1 = 15;
		} else c1 = 0;
		if (score[i+1] > scoreBase) {
			c2 = (score[i+1] - scoreBase) / scoreScaling;
			if (c2 > 15) c2 = 15;
		} else c2 = 0;

		thisScore[i/2] = c1 * 16 + c2;
	}

	if (i+1 == readLen) {
		uint8_t c1;
		if (score[i] > scoreBase) {
			c1 = (score[i] - scoreBase) / scoreScaling;
			if (c1 > 15) c1 = 15;
		} else c1 = 0;
		thisScore[i/2] = c1 * 16;
	}

	_next = next;
}


// Contains one read
class InputReadNode
{
public:
	InputReadNode(const ShortDnaBinarySequence* readSeq, ReadAndScoreType scoreType, const char* score);
	void addCopy(ReadAndScoreType scoreType, const char* score);

	~InputReadNode() { 
		while (_scores) {
			ScoreList* n = _scores->next();
			delete _scores;
			_scores = n;
		}
		if (_seq) SimpleAllocator::free((uint8_t*) _seq);
	}
	uint16_t length() const { return _seq->length(); }
	uint32_t numCopies() const { return _numCopies; }
	uint64_t memorySize() const;
	void write(ostream& o, uint16_t mLength, uint16_t readSize, uint16_t recordSize, uint32_t* scoreBuffer) const;

	void *operator new(size_t s) { return SimpleAllocator::allocate(s); }
	void operator delete(void *p) { SimpleAllocator::free((uint8_t*) p); }
	uint32_t depth() const {
		uint32_t ret = 0;
		if (_left) ret = _left->depth();
		if (_right) {
			uint32_t r = _right->depth();
			if (r > ret) ret = r;
		}
		return ret + 1;
	}

	friend class InputReadTree;

private:
	// no copying:
	InputReadNode(InputReadNode&);
	InputReadNode& operator=(InputReadNode&);

	InputReadNode *_left, *_right;
	ShortDnaBinarySequence* _seq;
	ScoreList* _scores;		// linked list of scores for this read
	uint32_t _numCopies;		// number of times this read occurs
};

static Stack<const InputReadNode *> nodeStack;	// shared by everyone, so use with care and leave empty

InputReadNode::InputReadNode(const ShortDnaBinarySequence* seq, ReadAndScoreType scoreType, const char* score)
	: _left(NULL), _right(NULL), _seq(NULL), _scores(NULL), _numCopies(1)
{
	uint16_t readLen = seq->length();
	_seq = new (readLen) ShortDnaBinarySequence(seq);

	if (score)
		_scores = new (readLen) ScoreList(scoreType, score, readLen, _scores);
}


void
InputReadNode::addCopy(ReadAndScoreType scoreType, const char* score)
{
	_numCopies++;
//	if (_numCopies == 10) cdbg << "qq10: " << _seq->toText() << endl;
	if (!score) return;

	uint16_t readLen = length();		// length of read in bases
	_scores = new (readLen) ScoreList(scoreType, score, readLen, _scores);
}


uint64_t
InputReadNode::memorySize() const
{
	uint64_t ret = 0;
	const InputReadNode* node = this; 
	while (node) {		// order of traveral doesn't matter here
		ret += sizeof(InputReadNode);
		if (node->_seq) {
			ret += node->_seq->memorySize();
			if (node->_scores)
				ret += node->_scores->memorySize(_seq->length());
		}
		if (node->_left) {
			if (node->_right) {
				nodeStack.push(node->_left);
				node = node->_right;
			} else {
				node = node->_left;
			}
		} else {
			node = node->_right;
			if (!node and !nodeStack.isEmpty())
				node = nodeStack.pop();
		}
	}
	assert(nodeStack.isEmpty());
	return ret;
}


void
InputReadNode::write(ostream& o, uint16_t mLength, uint16_t readSize, uint16_t recordSize, uint32_t* scoreBuffer) const
{
	static uint8_t outBuffer[lineLength];
	static uint8_t* outBuffP = outBuffer;	// to avoid warning
	memset(outBuffer, 0, recordSize);

	((ShortDnaBinarySequence*)outBuffP)->copy(_seq, mLength);
	uint8_t* next = outBuffer + readSize;

	* ((uint32_t*) next) = _numCopies;
	next += 4;

	if (useScores) {
		uint16_t len = _seq->length();
		if (len > mLength) len = mLength;
		uint16_t k;
		if (_numCopies > 1) {
			ScoreList* scores = _scores;
			for (k = 0; k < len; ++k)
				scoreBuffer[k] = scores->score(k);
			for (scores = scores->next(); scores; scores = scores->next()) {
				// Make the score of i'th read the average of all the reads with same sequence:

				for (k = 0; k < len; ++k)
					scoreBuffer[k] += scores->score(k);
			}
			for (k = 0; k + 1 < len; k += 2, ++next) {
				uint8_t c1 = (uint8_t) (scoreBuffer[k] / _numCopies);
				if (c1 > 15) c1 = 15;
				uint8_t c2 = (uint8_t) (scoreBuffer[k+1] / _numCopies);
				if (c2 > 15) c2 = 15;
				*next = c1 * 16 + c2;
			}
			if (k + 1 == len) {
				uint8_t c1 = (uint8_t) (scoreBuffer[k] / _numCopies);
				if (c1 > 15) c1 = 15;
				*next = c1 * 16;
				next++;
			}
		} else {
			uint16_t scoreLen = len/2;
			memcpy(next, _scores->rawScore(), scoreLen);
			if (len % 2) {
				next[scoreLen] = _scores->rawScore()[scoreLen] & 0xf0;
				scoreLen++;
			}
			next += scoreLen;
		}
	}
	o.write((char*)outBuffer, recordSize);
}


// Tree structure of reads
class InputReadTree
{
public:
	InputReadTree() : _root(NULL), _number(0), _numUnique(0), _readLength(0), _maxDepth(0) {}
	~InputReadTree();
	uint64_t memorySize() const { 
		uint64_t ret = sizeof(InputReadTree);
		if (_root) ret += _root->memorySize();
		return ret;
	}

	uint16_t add(char* textSeq, ReadAndScoreType scoreType, char* score = NULL);
	uint64_t number() const { return _number; }
	uint64_t numUnique() const { return _numUnique; }
	void write(ostream& o) const;
	uint16_t writeStats(ostream& o);
	uint64_t size() const;
	uint16_t readLength () const { return _readLength; }
	void setReadLength (uint16_t rl) { _readLength = rl; }
	uint64_t maxDepth() const { return _maxDepth; }
;
private:
	// no copying:
	InputReadTree(InputReadTree&);
	InputReadTree& operator=(InputReadTree&);

	void addNode(const char* textSeq, ReadAndScoreType scoreType, const char* score = NULL);

	InputReadNode* _root;
	uint64_t _number;
	uint64_t _numUnique;
	uint64_t _readLength;
	uint64_t _maxDepth;
};


InputReadTree::~InputReadTree()
{
	static Stack<InputReadNode *> stack;
	InputReadNode* node = _root;
	while (node) {
		if (node->_left) {
			stack.push(node);
			node = node->_left;
		} else {
			InputReadNode* r = node->_right;
			delete node;
			node = r;
		}
		while (!node and !stack.isEmpty()) {
			node = stack.pop();
			InputReadNode* r = node->_right;
			delete node;
			node = r;
		}
	}
}

 
void
InputReadTree::addNode(const char* seq, ReadAndScoreType scoreType, const char* score)
{
	static uint8_t seqBuff[lineLength];
	static ShortDnaBinarySequence * thisSeq = (ShortDnaBinarySequence*) seqBuff;

	_number++;
	
	thisSeq->set(seq);
	InputReadNode** nodeP = &_root;
	uint64_t depth = 0;

	while (true) {
		if (*nodeP == NULL) {
			*nodeP = new InputReadNode(thisSeq, scoreType, score);
			_numUnique++;
			depth++;
			if (depth > _maxDepth) _maxDepth = depth;
			return;
		}
		int cmp = thisSeq->compare((*nodeP)->_seq);
		if (cmp < 0) {
			nodeP = &((*nodeP)->_left);
			depth++;
		} else if (cmp > 0) { 
			nodeP = &((*nodeP)->_right);
			depth++;
		} else {
			(*nodeP)->addCopy(scoreType, score);
			return;
		}
	}
}


uint64_t 
InputReadTree::size() const
{
	uint64_t ret = 0;
	const InputReadNode* node = _root;
	while (node) {
		if (node->_left) {
			nodeStack.push(node);
			node = node->_left;
		} else {
			ret++;
			node = node->_right;
		}
		while (!node and !nodeStack.isEmpty()) {
			node = nodeStack.pop();
			ret++;
			node = node->_right;
		}
	}
	assert(nodeStack.isEmpty());
	return ret;
}


static uint16_t
cleanReadSeq(char*& seq)
{
	uint16_t len = strlen(seq);
	char* start;
	for (start = seq; *start and *start != 'A' and *start != 'C' and *start != 'G' and *start != 'T'
	and *start != 'a' and *start != 'c' and *start != 'g' and *start != 't'; ++start);

	char* end;
	for (end = start; 
	*end and (*end == 'A' or *end == 'C' or *end == 'G' or *end == 'T'
	or *end == 'a' or *end == 'c' or *end == 'g' or *end == 't');
	end++);

	if (end - start < minLength or end - start > maxLength)
		return 0;

	if (end - start == len)
		return len;

	seq = start;
	*end = 0;
	return end - start;
}


static uint16_t
cleanReadSeq(char*& seq, char*& score)
{
	const char* origSeq = seq;
	uint16_t len = cleanReadSeq(seq);
	if (seq > origSeq) score += seq - origSeq;
	score[len] = 0;
	return len;
}


uint16_t
InputReadTree::add(char* seq, ReadAndScoreType scoreType, char* score)
{
	if (!useScores) score = NULL;
	uint16_t len;

	if (score) len = cleanReadSeq(seq, score);
	else len = cleanReadSeq(seq);

	if (len < (uint16_t) minLength or len > (uint16_t) maxLength) return 0;

	addNode(seq, scoreType, score);

	static char revcom[lineLength];
	Util::reverseComplement(revcom, seq);

	if (score) {
		static char revScore[lineLength];
		uint16_t i;
		for (i = 0; i < len; ++i)
			revScore[i] = score[len-i-1];
		revScore[i] = '\0';
		addNode(revcom, scoreType, revScore);
	} else {
		addNode(revcom, scoreType);
	}

	if (len > _readLength) _readLength = len;
	return len;
}


void
InputReadTree::write(ostream& o) const
{
	uint16_t readSize = ShortDnaBinarySequence::memorySize(_readLength);
	uint16_t scoreSize = useScores ? (_readLength+1)/2: 0;
	uint16_t recordSize = readSize + 4 + scoreSize;
	uint32_t scoreBuffer[_readLength];

	// header line: signature version useScores numUniqueReads totalNumReads readLength memoryUsed localTime
	o << prinsesReadFileSignature << "\t" 
		<< prinsesReadFileVersion << "\t"
		<< (useScores ? "1" : "0") << "\t"
		<< _numUnique << "\t"
		<< _number << "\t"
		<< _readLength << "\t"
		<< (recordSize * _numUnique) << "\t"
		<< Util::getLocalTime(time(NULL)) << endl;
	
	const InputReadNode* node = _root;
	while (node) {
		if (node->_left) {
			nodeStack.push(node);
			node = node->_left;
			continue;
		}
		node->write(o, _readLength, readSize, recordSize, scoreBuffer);
		node = node->_right;
		while (!node and !nodeStack.isEmpty()) {
			node = nodeStack.pop();
			node->write(o, _readLength, readSize, recordSize, scoreBuffer);
			node = node->_right;
		}
	}
	assert(nodeStack.isEmpty());
}


uint16_t	// returns the read length of the 99'th percentile
InputReadTree::writeStats(ostream& o)
{
	uint32_t readLengthDist[1024];
	uint16_t percentile99 = 0;
	for (uint16_t i = 0; i < 1024; ++i)
		readLengthDist[i] = 0;
	
	const InputReadNode* node = _root;
	while (node) {
		if (node->_left) {
			nodeStack.push(node);
			node = node->_left;
			continue;
		}

		// nothing on the left, process this node:
		uint16_t len = node->length();
		if (len < 1024) readLengthDist[len] += node->numCopies();

		// move on to the right:
		node = node->_right;

		while (!node and !nodeStack.isEmpty()) {
			node = nodeStack.pop();

			// we've already done the left side, process this node:
			uint16_t len = node->length();
			if (len < 1024) readLengthDist[len] += node->numCopies();

			node = node->_right;
		}
	}
	assert(nodeStack.isEmpty());

	uint64_t cummulative = 0;
	o << "Read length\tNumber of reads\n";
	for (uint16_t i = 0; i < 1024; ++i) {
		if (readLengthDist[i]) {
			o << setw(4) << i << "\t" << setw(12) << readLengthDist[i] << endl;
			cummulative += readLengthDist[i];
			if (!percentile99 and cummulative * 100 > _number * 99)
				percentile99 = i;
		}
	}
	o << "Total:\t" << setw(12) << _number << endl << _numUnique << " unique reads, tree depth " << _maxDepth << endl;
	return percentile99;
}


static int16_t
getLine(gzFile fh, char* lineBuffer)
{
	if (gzgets(fh, lineBuffer, lineLength) == NULL) return -1;
	uint64_t l = strlen(lineBuffer);
	if (l > 0 and lineBuffer[l-1] == '\n') {
		--l;
		lineBuffer[l] = '\0';
	}
	return l;
}


static void
loadFastaReadsFile(gzFile& in, InputReadTree& reads)
{
	static char line[lineLength];
	String seq;
	int16_t lineLen;
	uint16_t readLength = 0, thisReadLength;
	uint64_t numReads = 0;

	while ((lineLen = getLine(in, line)) >= 0) {
		if (numReads % 1000000 == 0) cerr << '.';
		if (line[0] == '>') {
			if (!seq.empty() and seq.length() < lineLength) {
				strcpy(line, seq.c_str());
				thisReadLength = reads.add(line, Fasta);
				if (thisReadLength) {
					if (readLength < thisReadLength) readLength = thisReadLength;
					numReads++;
				}
			}
			seq.clear();
		} else {
			seq += line;
		}
	}
	if (!seq.empty() and seq.length() < lineLength) {
		strcpy(line, seq.c_str());
		thisReadLength = reads.add(line, Fasta);
		if (thisReadLength) {
			if (readLength < thisReadLength) readLength = thisReadLength;
			numReads++;
		}
	}
	cerr << " " << numReads << " reads, length=" << readLength << ", cummulative " << (reads.number()/1000000) << "M reads, memory usage: " << setprecision(3) << Util::getMemoryUsageGB() << "GB\n";
}


static int16_t
averageScore(const char* score, ReadAndScoreType scoreType)
{
	int32_t sumScore = 0, i;
	char scoreBase = qualityScoreBase[scoreType];
	for (i = 0; score[i]; ++i)
		sumScore += score[i] - scoreBase;
	return i ? (int16_t) (sumScore / i) : 0;
}


static void
loadFastqReads(gzFile& in, InputReadTree& reads, ReadAndScoreType scoreType)
{
	static char line[lineLength];
	uint64_t numReads = 0;
	uint64_t lineNo = 0;
	uint16_t readLength = 0;
	int16_t lineLen;
	
	for (; (lineLen = getLine(in, line)) >= 0; ++numReads) {
		lineNo++;
		if (numReads % 1000000 == 0) cerr << '.';
		if (lineLen < 1 or line[0] != '@') {
			cerr << programName << ": file not in fastq format, line " << lineNo
	#ifdef DEBUG
				<< " in " << __PRETTY_FUNCTION__ << " " << __FILE__ 
				<< " line " << __LINE__ 
	#endif
				<< endl;
			exit(1);
		}
		uint16_t thisReadLength = getLine(in, readSequence); lineNo++;
		lineLen = getLine(in, line); lineNo++;
		if (lineLen < 1 or line[0] != '+') {
			cerr << programName << ": file not in fastq format, line " << lineNo
	#ifdef DEBUG
				<< " in " << __PRETTY_FUNCTION__ << " " << __FILE__ 
				<< " line " << __LINE__ 
	#endif
				<< endl;
			exit(1);
		}
		uint16_t scoreLength = getLine(in, line); lineNo++;
		char* score = line;
		if (thisReadLength != scoreLength) {
			cerr << programName << ": invalid fastq file - length of score is different from length of sequence (" << readSequence << " vs " << score
	#ifdef DEBUG
				<< " in " << __PRETTY_FUNCTION__ << " " << __FILE__ 
				<< " line " << __LINE__ 
	#endif
				<< endl;
			exit(1);
		}

		if (thisReadLength < (uint32_t) minLength or thisReadLength > (uint32_t) maxLength) continue;
		if (minAverageScore and scoreLength and averageScore(score, scoreType) < minAverageScore)
			continue;

		thisReadLength = reads.add(readSequence, scoreType, score);
		if (readLength < thisReadLength) readLength = thisReadLength;
	}
	cerr << " " << numReads << " reads, length=" << readLength << ", cummulative " << (reads.number()/1000000) << "M reads, memory usage: " << setprecision(3) << Util::getMemoryUsageGB() << "GB\n";
}
 
static const char samBaseMap[16] = { 0, 'A', 'C', 0, 'G', 0, 0, 0, 'T', 0, 0, 0, 0, 0, 0, 0 };

static void
loadBamReads(const char* filename, InputReadTree& reads)
{
	samfile_t* sfp = samopen(filename, "rb", 0);
	if (sfp == NULL) {
		cerr << "Can't open file " << filename << endl;
		return;
	}
	cerr << "Loading bam file " << filename << ' ';

	uint64_t numReads = 0, numReadsAdded = 0, numReadsThatMapToTooManyLoci = 0;
	uint64_t numReadsTooShort = 0, numReadsTooLong = 0, numReadsLowQuality = 0;
	uint64_t numReadsUnmapped = 0;
	bam1_t* alignment = bam_init1();
	int32_t maxQual = 0;
	int32_t minQual = 10000;
	uint16_t readLength = 0;
	while (samread(sfp, alignment) > 0) {
		numReads++;
		if (numReads % 1000000 == 0) cerr << '.';
		uint16_t thisReadLength = alignment->core.l_qseq;


		if (thisReadLength < (uint16_t) minLength) {
			numReadsTooShort++;
			continue;
		}
		if (thisReadLength > (uint16_t) maxLength) {
			numReadsTooLong++;
			continue;
		}

		// this is BWA-specific: exclude sequences that map to too many loci
		uint8_t* s = bam_aux_get(alignment, "X0");
		int32_t numMappings = s ? bam_aux2i(s) : 0;
		if (numMappings > maxRepeats) {
			numReadsThatMapToTooManyLoci++;
			continue;
		}
/*
		s = bam_aux_get(alignment, "X1");
		numMappings += s ? bam_aux2i(s) : 0;
		if (numMappings > maxRepeats) {
			numReadsThatMapToTooManyLoci++;
			continue;
		}

*/

		uint8_t* seq = bam1_seq(alignment);
		uint8_t* qual = NULL;
		if (useScores or minAverageScore) {
			qual = bam1_qual(alignment);
			for (int32_t i = 0; i < thisReadLength; ++i) {
				char base = samBaseMap[bam1_seqi(seq, i)];
				if (base == 0) thisReadLength = i;
				readSequence[i] = base;
				if (qual[i] > maxQual) maxQual = qual[i];
				if (qual[i] < maxQual) minQual = qual[i];
			}
		} else {
			for (int32_t i = 0; i < thisReadLength; ++i) {
				char base = samBaseMap[bam1_seqi(seq, i)];
				if (base == 0) thisReadLength = i;
				readSequence[i] = base;
			}
		}
		readSequence[thisReadLength] = 0;
		if (thisReadLength < (uint16_t) minLength) {
			numReadsTooShort++;
			continue;
		}

		if (minAverageScore and averageScore((char*) qual, Raw) < minAverageScore) {
			numReadsLowQuality++;
			continue;
		}
		thisReadLength = reads.add(readSequence, Raw, (char*) qual);
		if (readLength < thisReadLength) readLength = thisReadLength;
		if (thisReadLength) {
			numReadsAdded++;
			if (alignment->core.flag & BAM_FUNMAP) numReadsUnmapped++;
		} else
			numReadsTooShort++;

	}
	bam_destroy1(alignment);
	samclose(sfp);
	cerr	<< endl
		<< "Reads file                               " << filename << endl
		<< "Read length                              " << readLength << endl
		<< "Best base quality                        " << maxQual << endl
		<< "Worst base quality                       " << minQual << endl
		<< "Total number of reads                    " << numReads << endl;
	if (numReadsThatMapToTooManyLoci)
	cerr	<< "Reads that are mapped to more than " << maxRepeats << " loci " << numReadsThatMapToTooManyLoci << endl;
	if (numReadsTooShort)
	cerr	<< "Reads that are shorter than minLength    " << numReadsTooShort << endl;
	if (numReadsTooLong)
	cerr	<< "Reads that are longer than maxLength     " << numReadsTooLong << endl;
	if (numReadsLowQuality)
	cerr	<< "Reads of low quality                     " << numReadsLowQuality << endl;
	cerr	<< "Number of reads used                     " << numReadsAdded << endl
		<< " of which unmapped                       " << numReadsUnmapped;
	if (numReadsAdded) cerr << " (" << 100 * numReadsUnmapped / numReadsAdded << "%)";
	cerr << endl
		<< "Cummulative (incl reverse complement)    " << reads.number()/1000000 << "M reads" << endl
		<< "Memory usage                             " << setprecision(3) << Util::getMemoryUsageGB() << "GB" << endl
		<< endl;
}


static void
doIt(int argc, char* argv[])
{
	stringstream signature;
	signature << argv[0] << ' ' << argv[1];
	programName = signature.str();
	argc--; argv++;
	for (int i = 1; i < argc; ++i)
		signature << ' ' << argv[i];
#ifdef DEBUG
	signature << " debug";
#endif

	time_t time0 = time(NULL);
	cerr << Util::getLocalTime(time0) << " " << signature.str() << endl;

	// Get command line arguments:
	map<String, bool*> flags;
	map<String, int*> intOptions;
	map<String, String*> stringOptions;
	flags["scores"] = &useScores;
	flags["version"] = &printVersion;
	flags["discardTooLong"] = &discardTooLong;
	intOptions["minLength"] = &minLength;
	intOptions["maxLength"] = &maxLength;
	intOptions["filter"] = &minAverageScore;
	vector<String> argV = Util::getCommandLineOptions(flags, intOptions, stringOptions, argc, argv, usage);
	if (printVersion) {
		cerr << "built on " << __DATE__ << " " << __TIME__ << endl;
		exit(0);
	}
	if (argV.size() < 1) usage();

	InputReadTree reads;

	if (oldVersion) {
		// Command line arguments:
		String typeString = argV[0];
		ReadAndScoreType scoreType;
		if (typeString == "illumina") {
			scoreType = Illumina;
		} else if (typeString == "phred") {
			scoreType = Phred;
		} else if (typeString == "fasta") {
			scoreType = Fasta;
			useScores = false;
		} else {
			usage();
		}

		// See if we can open all files first:
		vector<gzFile> inFH;
		vector<String> filenames;
		vector<ReadAndScoreType> scoreTypes;
		for (uint64_t i = 1; i < argV.size(); ++i) {
			if (argV[i] == "illumina") {
				scoreType = Illumina;
				continue;
			}
			if (argV[i] == "phred") {
				scoreType = Phred;
				continue;
			}
			if (argV[i] == "fasta") {
				scoreType = Fasta;
				useScores = false;
				continue;
			}
			const char* fileName = argV[i].c_str();
			gzFile in = gzopen(fileName, "r");
			if (in == NULL) {
				cerr << programName << ": Can't open file " << fileName
		#ifdef DEBUG
					<< " in " << __PRETTY_FUNCTION__ << " " << __FILE__ 
					<< " line " << __LINE__ 
		#endif
					<< endl;
				exit(1);
			}
			inFH.push_back(in);
			filenames.push_back(argV[i]);
			scoreTypes.push_back(scoreType);
		}

		// Load fasta/fastq files:
		uint16_t numberReadsFiles = inFH.size();
		for (uint16_t i = 0; i < numberReadsFiles; ++i) {
			cerr << "Loading file " << filenames[i] << ": (" << (i+1) << " of " << numberReadsFiles << ") ";
			if (scoreTypes[i] == Fasta)
				loadFastaReadsFile(inFH[i], reads);
			else
				loadFastqReads(inFH[i], reads, scoreTypes[i]);
			gzclose(inFH[i]);
		}
	}

	for (uint64_t i = 0; i < argV.size(); ++i)
		loadBamReads(argV[i].c_str(), reads);

	uint16_t percentile99 = reads.writeStats(cerr);
	if (discardTooLong and percentile99 != reads.readLength()) {
		cerr << "Truncating reads over " << percentile99 << " bp\n";
		reads.setReadLength(percentile99);
	}

	cerr << "Memory usage: " << setprecision(3) << Util::getMemoryUsageGB() << " GB (calculated = " << setprecision(3) << reads.memorySize() << ", actual = " << Util::getMemoryUsage() << ")\n";

	cerr << Util::getLocalTime(time(NULL)) << " writing...\n";
	reads.write(cout);

	// Wind down:
	time_t time1 = time(NULL);
	cerr << Util::getLocalTime(time1) << " " << programName << " done in " << Util::hms(time1-time0) << endl;
}


void
main(int argc, char* argv[])
{
	Util::setThousandSeparator(cerr); // shouldn't do the same for stdout, as it's the reads file
	SimpleAllocator::initialise();
	PrinsesPrepare::doIt(argc, argv);
	SimpleAllocator::freeAll();
}

}; // namespace PrinsesPrepare
