#!/usr/bin/perl

# Copyright 2009 Andreas Massouras
#
# This file is part of Primer-Initiated Sequence Synthesis (PrInSeS)
#
# For license information, please refer to the LICENSE document included in the distribution.
# By downloading or using this software, you accept and are bound by the terms of the license.

package Variants;

use strict;
use warnings;
use Carp;

sub max { $_[0] > $_[1] ? $_[0] : $_[1]; }

# Reads file and returns variants in structure: $variants->{seqName} = [ position (starting from 0), replaced seq length, new sequence, strand, text&comments ]
sub loadVariants
{
	my ($variantFile) = @_;
	my $variants = {};
	my $text = '';
	my @header;
	my $last = \@header;

	open(VARS, $variantFile) || confess ("Can't read $variantFile: $!");
	while (my $l = <VARS>) {
		if ($l =~ /^#/) {
			$text .= $l;
			next;
		}
		push(@$last, $text) if ($text);
		$text = $l;
		chomp($l);
		my ($seqName, $pos, $oldSeqLength, $newSeq) = split(/\t/, $l);
		$pos--;
		my $strand = '';
		if ($seqName =~ /^(.*)\:([udb])$/) {
			$seqName = $1;
			$strand = $2;
		}
		$variants->{$seqName} = [] unless(defined($variants->{$seqName}));
		push(@{$variants->{$seqName}}, $last = [ $pos, $oldSeqLength, $newSeq, $strand ]);
	}
	push(@$last, $text) if ($text);
	close(VARS);

	my $header = @header ? $header[0] : '';
	($variants, $header);
}


sub compareVariants
{
	# $variantsA and $variantsB must point to arrays of variants
	# Each variant is an array of 3 elements: [position (STARTING AT 0), length of deleted seq, inserted seq]
	# Returns copy of $variantsA and $variantsB
	my ($labelA, $refseqA, $variantsA, $labelB, $refseqB, $variantsB) = @_;

	my ($inBothIns, $inBothDel, $inBothOther) = (0,0,0);
	my $overlapping = 0;
	if (not $variantsA or not $variantsB) {
		return ($inBothIns, $inBothDel, $inBothOther, $overlapping, $variantsA, $variantsB);
	}

	# Take a copy of variants arrays:
	{
		my @ia = sort { $a->[0] <=> $b->[0] } @$variantsA;
		my @ib = sort { $a->[0] <=> $b->[0] } @$variantsB;
		$variantsA = [];
		$variantsB = [];
		for (my $i = 0; $i < @ia; ++$i) {
			push(@$variantsA, [ @{$ia[$i]} ]);
		}
		for (my $i = 0; $i < @ib; ++$i) {
			push(@$variantsB, [ @{$ib[$i]} ]);
		}
	}
VARIANT:
	for (my $ia = 0; $ia < @$variantsA; ++$ia) {
		my ($aPos, $aOldSeqLength, $aNewSeq) = @{$variantsA->[$ia]};
		for (my $ib = 0; $ib < @$variantsB; ++$ib) {
			my ($bPos, $bOldSeqLength, $bNewSeq) = @{$variantsB->[$ib]};
			next if (abs($aPos - $bPos) > 2000);
			if ($aOldSeqLength != $bOldSeqLength or length($aNewSeq) != length($bNewSeq)) {
				if (($aPos <= $bPos and $aPos + max($aOldSeqLength, length($aNewSeq)) >= $bPos)
				or ($bPos <= $aPos and $bPos + max($bOldSeqLength, length($bNewSeq)) >= $aPos )) {
					$overlapping++;
					$variantsA->[$ia][4] .= "# variantCompare: Overlaps with $labelB (".($bPos+1).")\n";
					$variantsB->[$ib][4] .= "# variantCompare: Overlaps with $labelA (".($aPos+1).")\n";
				}
				next;
			}
			my $found = 0;

			if ($aPos == $bPos) {
				$found = 1 if ($aNewSeq eq $bNewSeq);
			} elsif ($aPos < $bPos) {
				my $aSeq = $aNewSeq . substr($refseqA, $aPos + $aOldSeqLength, $bPos - $aPos);
				my $bSeq = substr($refseqB, $aPos, $bPos - $aPos) . $bNewSeq;
				$found = 1 if ($aSeq eq $bSeq);
			} else { # $aPos > $bPos
				my $bSeq = $bNewSeq . substr($refseqB, $bPos + $bOldSeqLength, $aPos - $bPos);
				my $aSeq = substr($refseqA, $bPos, $aPos - $bPos) . $aNewSeq;
				$found = 1 if ($aSeq eq $bSeq);
			}
			if ($found) {
				$variantsA->[$ia][4] .= "# variantCompare: Same as $labelB (".($bPos+1).")\n";
				$variantsB->[$ib][4] .= "# variantCompare: Same as $labelA (".($aPos+1).")\n";
				if ($aOldSeqLength == 0) {
					++$inBothIns;
				} elsif (length($aNewSeq) == 0) {
					++$inBothDel;
				} else {
					++$inBothOther;
				}
				next VARIANT;
			} else {
				if (($aPos <= $bPos and $aPos + $aOldSeqLength - length($aNewSeq) > $bPos)
				or ($bPos <= $aPos and $bPos + $bOldSeqLength - length($bNewSeq) > $aPos )) {
					$overlapping++;
					$variantsA->[$ia][4] .= "# variantCompare: Overlaps with $labelB (".($bPos+1).")\n";
					$variantsB->[$ib][4] .= "# variantCompare: Overlaps with $labelA (".($aPos+1).")\n";
				}
			}
		}
	}

	($inBothIns, $inBothDel, $inBothOther, $overlapping, $variantsA, $variantsB);
}

1;
