/*

Copyright 2010 Andreas Massouras

This file is part of Primer-Initiated Sequence Synthesis (PrInSeS)

For license information, please refer to the LICENSE document included in the distribution.
By downloading or using this software, you accept and are bound by the terms of the license.

*/

#include <fstream>
#include <zlib.h>
#include "Genome.h"
#include "Variant.h"

const Strand Strand::positive(NULL, '+');
const Strand Strand::negative(NULL, '-');

const ChromosomeIndex invalidChromosomeIndex = numeric_limits<ChromosomeIndex>::max();

// sort chromosomes by their name:
static bool lessThan(DnaNamedSequence* const & a, DnaNamedSequence* const & b)
{
	return a->name() < b->name();
}


void
Genome::setFromVector(const Vector<DnaNamedSequence*>& seqs)
{
	if (seqs.size() > 65535) {
		cerr << "Max number of sequences for genome exceeded\n";
		exit(-1);
	}

	VectorI o(seqs.order(lessThan));

	_chromosomeNames.resize(seqs.size());
	_chromosomeSequenceStarts.resize(seqs.size());
	_chromosomeSequenceLengths.resize(seqs.size());

	// Find out how many non-nucleotide bases (Ns) are at both ends of each chromosome:
	for (ChromosomeIndex i = 0; i < seqs.size(); ++i) {
		const DnaNamedSequence& thisChromo = *seqs[o[i]];
		_chromosomeNames[i] = thisChromo.name();
		const GenomicPosition len = thisChromo.length();
		_chromosomeSequenceStarts[i] = _length;	// so far
		_chromosomeSequenceLengths[i] = len;
		_length += len;
	}

	_sequence = new char [_length + 1];
	_sequenceReverseComplement = new char [_length + 1];
	char *s = _sequence;

	// check genome alphabet: we should only have ACGTN
	for (ChromosomeIndex i = 0; i < seqs.size(); i++) {
		DnaNamedSequence& thisChromo = *seqs[o[i]];
		const char* seq = thisChromo.c_str();
		GenomicPosition len = _chromosomeSequenceLengths[i];
		for (; len; --len, ++s, ++seq) {
			switch(*seq) {
			case 'A': case 'C': case 'G': case 'T':  case 'N' :
				*s = *seq;
				break;
			case 'a': 
				*s = 'A';
				break;
			case 'c':
				*s = 'C';
				break;
			case 'g':
				*s = 'G';
				break;
			case 't': 
				*s = 'T';
				break;
			default:
				*s = 'N';
			}
		}
	
	}
	*s = 0;

	Util::reverseComplement(_sequenceReverseComplement, _sequence);
}


Genome::Genome(const char* filename) 
	: _filename(filename), _sequence(NULL), _sequenceReverseComplement(NULL), _length(0)
{
	Vector<DnaNamedSequence*> seqs;
	loadFastaFile(filename, seqs);
	setFromVector(seqs);
	seqs.freeAll();
}


Genome::Genome(const Vector<DnaNamedSequence*>& seqs)
	: _sequence(NULL), _sequenceReverseComplement(NULL), _length(0)
{
	setFromVector(seqs);
}


uint32_t
Genome::checksum() const
{
	uint32_t ret = adler32(0L, Z_NULL, 0);
//	for (ChromosomeIndex i = 0; i < _sequences.size(); i++) {
		ret = adler32(ret, (const Bytef*) _sequence, _length);
//	}
	return ret;
}


void
NamedGenomicRegion::loadBEDfile(const char* filename, const Genome& genome, bool firstThreeColumnsOnly, /* returned: */ list<NamedGenomicRegion>& regions)
{
// see http://genome.ucsc.edu/FAQ/FAQformat.html#format1
	ifstream ifile(filename, ios_base::in);
	if (ifile.fail()) {
		cerr << ": Can't open file " << filename << " in "
			<< __PRETTY_FUNCTION__ << " " << __FILE__ 
			<< " line " << __LINE__ << endl;
		exit(1);
	}

	String line;
	String chromoS, startS, endS, id, score, strandS;
	String *fields[] = { &chromoS, &startS, &endS, &id, &score, &strandS, NULL };
	String *fields3[] = { &chromoS, &startS, &endS, NULL };
	uint32_t lineNo = 0;
	map<String, uint32_t> sequencesNotFound;
	while (getline(ifile, line)) {
		lineNo++;
		if (line[0] == '#') continue;
		if (firstThreeColumnsOnly) {
			if (Util::split(line, fields3, '\t') < 3) goto BOMB_OUT;
		} else {
			if (Util::split(line, fields, '\t') < 6) goto BOMB_OUT;
		}
		if (chromoS.length() < 4 or chromoS.substr(0 ,3) != "chr") continue;
		ChromosomeIndex index = genome.chromoIndex(chromoS.substr(3));
		if (index == Genome::SequenceNotFound) {
			String chromo = chromoS.substr(3);
			map<String, uint32_t>::iterator i = sequencesNotFound.find(chromo);
			if (i == sequencesNotFound.end())
				sequencesNotFound[chromo] = 1;
			else
				i->second ++;
			continue;
		}
		GenomicPosition start, end;
		if (!Util::value<GenomicPosition>(start, startS)) goto BOMB_OUT;
		if (!Util::value<GenomicPosition>(end, endS)) goto BOMB_OUT;
		if (end < start) goto BOMB_OUT;
		if (end > genome.chromoLength(index)) {
			cerr << ": position " << end << " outside chromosome " << filename << " line " << lineNo << " in "
				<< __PRETTY_FUNCTION__ << " " << __FILE__ 
				<< " line " << __LINE__ << endl;
			exit(1);
		}
		GenomicPosition length = end - start;

		Strand strand = Strand::positive;
		if (!firstThreeColumnsOnly) {
			if (strandS == "-") {
				strand = Strand::negative;
			} else if (strandS != "+") {
				goto BOMB_OUT;
			}
		}

		const GenomicPosition pos = genome.position(index, start);
		if (pos != invalidGenomicPosition)
			regions.push_back(NamedGenomicRegion(id, pos, length, strand));
	}
	ifile.close();
	if (sequencesNotFound.size()) {
		map<String, uint32_t>::iterator i;
		cerr << "Warning: The following sequences in file " << filename << " were not found in the genome: ";
		for (i = sequencesNotFound.begin(); i != sequencesNotFound.end(); ++i)
			cerr << i->first << '(' << i->second << ") ";
		cerr << endl;
	}

	return;

BOMB_OUT:
	cerr << ": incorrect BED file format in " << filename << " line " << lineNo << " in "
		<< __PRETTY_FUNCTION__ << " " << __FILE__ 
		<< " line " << __LINE__ << endl;
	exit(1);
}


void
Transcript::loadUCSCtranscripts(const char* filename, const Genome& genome, /* return */ Vector<Transcript*> & transcripts)
{
	ifstream ifile(filename, ios_base::in);
	if (ifile.fail()) {
		cerr << ": Can't open file " << filename << " in "
			<< __PRETTY_FUNCTION__ << " " << __FILE__ 
			<< " line " << __LINE__ << endl;
		exit(1);
	}

//#bin    name    chrom   strand  txStart txEnd   cdsStart        cdsEnd  exonCount       exonStarts      exonEnds        id      name2   cdsStartStat    cdsEndStat      exonFrames
//73      NM_079882       chr4    +       380602  395607  381316  393018  13      380602,381212,382929,383296,383571,384119,384493,386299,390132,390435,391716,392438,392875,     380910,381632,383239,383508,383760,384265,385567,387388,390370,390590,391983,392802,395607,     187910  Hcf     cmpl    cmpl    -1,0,1,2,1,1,0,0,0,1,0,0,1,
	uint32_t lineNo = 0;
	list<Transcript*> tl;
	String line;
	Vector<String> fields;
	const char* errorMessage = NULL;
	Vector<String> exonStarts;
	Vector<String> exonEnds;
	map<String, Uint32_t> unknownChromo;
	while (getline(ifile, line)) {
		lineNo++;
		if (line[0] == '#') continue;
		Util::split(line, '\t', fields);
		if (fields.size() < 13) {
			errorMessage = "Invalid file format - too few fields ";
			break;
		}
		String chromo = fields[2];
		if (chromo.length() > 3 and chromo[0] == 'c' and chromo[1] == 'h' and chromo[2] == 'r')
			chromo = chromo.substr(3);
		ChromosomeIndex chromoIndex = genome.chromoIndex(chromo);
		if (chromoIndex == Genome::SequenceNotFound) {
			unknownChromo[chromo]++;
			continue;
		}

		String strandStr = fields[3];
		if (strandStr.length() == 0 or (strandStr[0] != '+' and strandStr[0] != '-')) {
			errorMessage = "Invalid format (strand) ";
			break;
		}
		Strand strand(NULL, strandStr[0]);
		GenomicPosition tss, tes, cdsStart, cdsEnd;
		if (!Util::valueU(tss, fields[4])) {
			errorMessage = "Invalid position for TSS";
			break;
		}
		tss--;
		if (!Util::valueU(tes, fields[5])) {
			errorMessage = "Invalid position for TES";
			break;
		}
		tes--;
		if (tes < tss) {
			GenomicPosition tmp = tes;
			tes = tss;
			tss = tmp;
		}
		if (tes >= genome.chromoLength(chromoIndex)) {
			errorMessage = "Transcript goes beyond the end of the chromosome";
			break;
		}

		if (!Util::valueU(cdsStart, fields[6])) {
			errorMessage = "Invalid position for CDS start";
			break;
		}
		cdsStart--;
		if (!Util::valueU(cdsEnd, fields[7])) {
			errorMessage = "Invalid position for CDS end";
			break;
		}
		cdsEnd--;
		if (cdsEnd < cdsStart) {
			GenomicPosition tmp = cdsEnd;
			cdsEnd = cdsStart;
			cdsStart = tmp;
		}
		if (cdsEnd > tes) {
			errorMessage = "Coding region goes beyond TES";
			break;
		}
		Transcript *newTranscript = new Transcript(fields[1],
			GenomicRegion(genome.position(chromoIndex, cdsStart), cdsEnd + 1 - cdsStart, strand),
			genome.position(chromoIndex, tss), tes + 1 - tss);

		Util::split(fields[9], ',', exonStarts);
		Util::split(fields[10], ',', exonEnds);
		if (exonStarts.size() != exonEnds.size()) {
			errorMessage = "Different number of exonStarts and exonEnds";
			break;
		}
		if (exonStarts.size() == 0) {
			errorMessage = "No exons defined";
			break;
		}
		uint16_t numExons = exonStarts.size();
		if (exonStarts[numExons-1].empty()) numExons--;
		newTranscript->_exons.resize(numExons);
		for (uint16_t i = 0; i < numExons; ++i) {
			GenomicPosition exonStart, exonEnd;
			if (!Util::valueU(exonStart, exonStarts[i])) {
				errorMessage = "Invalid exon start";
				break;
			}
//			exonStart--;
			if (!Util::valueU(exonEnd, exonEnds[i])) {
				errorMessage = "Invalid exon end";
				break;
			}
			exonEnd--;
			if (exonEnd < exonStart) {
				GenomicPosition tmp = exonEnd;
				exonEnd = exonStart;
				exonStart = tmp;
			}
			newTranscript->_exons[i] = Exon(genome.position(chromoIndex, exonStart), exonEnd + 1 - exonStart, strand);
		}
		if (errorMessage) break;
		tl.push_back(newTranscript);
	}

	if (errorMessage) {
		cerr << errorMessage << " file " << filename << " line " << lineNo << endl;
		exit(-1);
	}

	if (unknownChromo.size()) {
		cerr << "Warning: the following chromosomes in file " << filename << " are unknown and transcript data for them is ignored (number of transcripts): ";
		map<String, Uint32_t>::const_iterator c;
		for (c = unknownChromo.begin(); c != unknownChromo.end(); c++) {
			cerr << c->first << "(" << c->second << ") ";
		}
		cerr << endl;
	}

	transcripts.resize(tl.size());
	list<Transcript*>::iterator i;
	uint32_t j;
	for (i = tl.begin(), j = 0; i != tl.end(); ++i, ++j) {
		transcripts[j] = *i;
	}
}


void
Genome::applyVariants(const VariantSet& v, char direction)
{
	if (numberChromosomes() == 0 or v.size() == 0) return;

	list<Variant> variants(v);
	variants.sort();

	list<Variant>::iterator it;
	GenomicPosition newLength = _length;
	GenomicPosition lastPos = 0;
	for (it = variants.begin(); it != variants.end(); ++it) {
		if (direction == 'u' and it->_direction == 'd') continue;
		if (direction == 'd' and it->_direction == 'u') continue;
		const GenomicPosition pos = it->position();
		if (pos < lastPos) {
			cerr << "Genome " << _filename << ": Warning: variant at ";
			output(cerr, pos);
			cerr << " overlaps with previous one - ignored\n";
			continue;
		}
		if (pos > _length) {
			cerr << "Genome " << _filename
				<< ": Warning: variant with coordinate "
				<< (pos+1) << " is after the end of the genome - ignored\n";
			continue;
		}
		newLength += it->newSeqLength();
		GenomicPosition osl = it->oldSeqLength();
		newLength -= osl;
		lastPos = pos + osl;
	}
	char* newSequence = new char[newLength + 1];
	char* ns = newSequence;
	char* s = _sequence;

	lastPos = 0;
	ChromosomeIndex currentChromo = 0;
	GenomicPosition nextStart = _chromosomeSequenceLengths[0];
	Vector<GenomicPosition> newChromoLengths(_chromosomeSequenceLengths);
	int32_t runningOffset = 0;
	for (it = variants.begin(); it != variants.end(); ++it) {
		if (direction == 'u' and it->_direction == 'd') continue;
		if (direction == 'd' and it->_direction == 'u') continue;
		const GenomicPosition pos = it->position();
		if (pos < lastPos or pos > _length) continue;

		while (pos >= nextStart) {
			newChromoLengths[currentChromo] += runningOffset;
			runningOffset = 0;
			currentChromo++;
			assert(currentChromo < numberChromosomes());
			nextStart += _chromosomeSequenceLengths[currentChromo];
		}

		// copy original sequence before current variant:
		for (; lastPos < pos; lastPos++, ns++, s++)
			*ns = *s;

		GenomicPosition nsl = it->newSeqLength();
		GenomicPosition osl = it->oldSeqLength();
		runningOffset += nsl;
		runningOffset -= osl;

		// skip (possible) deletion:
		lastPos += osl;
		s += osl;

		if (!nsl) continue;

		// inserted sequence:
		const char *ins = it->newSeq().c_str();
		for (; nsl; nsl--, ns++, ins++)
			*ns = *ins;
	}
	newChromoLengths[currentChromo] += runningOffset;

	// copy what's left:
	for (; lastPos < _length; lastPos++, ns++, s++)
		*ns = *s;

	assert(s - _sequence == _length);
	assert(ns - newSequence == newLength);
	*ns = 0;
	delete [] _sequence;
	_sequence = newSequence;
	if (newLength > _length) {
		delete [] _sequenceReverseComplement;
		_sequenceReverseComplement = new char[newLength + 1];
	}
	Util::reverseComplement(_sequenceReverseComplement, _sequence);
	_length = newLength;

	_chromosomeSequenceLengths = newChromoLengths;
	GenomicPosition size = 0;
	for (ChromosomeIndex i = 0; i < newChromoLengths.size(); ++i) {
		_chromosomeSequenceStarts[i] = size;
		size += newChromoLengths[i];
	}
}


void 
Genome::output(ostream& o) const
{
	const char* seq = _sequence;
	for (ChromosomeIndex i = 0; i < _chromosomeNames.size(); ++i) {
		o << ">" << _chromosomeNames[i] << endl;

		GenomicPosition j = _chromosomeSequenceLengths[i];
		for (; j >= 80; j -= 80, seq += 80) {
			o.write(seq, 80);
			o << endl;
		}

		if (j) {
			o.write(seq, j);
			seq += j;
			o << endl;
		}
	}
	assert(seq - _sequence == _length);
}


void
Genome::output(const char* filename) const
{
	_filename = filename;
	ofstream fh(filename);
	if (fh.fail()) {
		cerr << "Can't open file " << filename << " for writing"
#ifdef DEBUG
			<< " in " << __PRETTY_FUNCTION__ 
			<< " " << __FILE__ << " line " << __LINE__ 
#endif
			<< endl;
		exit(1);
	}
	output(fh);
}


Genome::Genome(const Genome& other)
	: _filename(other._filename), _chromosomeNames(other._chromosomeNames), _chromosomeSequenceStarts(other._chromosomeSequenceStarts),
	_chromosomeSequenceLengths(other._chromosomeSequenceLengths),
	_sequence(new char[other._length+1]), _sequenceReverseComplement(new char[other._length+1]),
	_length(other._length)
{
	strncpy(_sequence, other._sequence, _length+1);
	strncpy(_sequenceReverseComplement, other._sequenceReverseComplement, _length+1);
}


void
Genome::output(ostream& o, const Variant& v) const
{
	GenomicPosition chromoPos = invalidGenomicPosition;
	o << chromoNamePos(v.position(), chromoPos) << ':' << v._direction;
	o << '\t' << chromoPos + 1
		<< '\t' << v.oldSeqLength()
		<< '\t' << v.newSeq()
		<< endl << v.text();
}
