/*

Copyright 2010 Andreas Massouras

This file is part of Primer-Initiated Sequence Synthesis (PrInSeS)

For license information, please refer to the LICENSE document included in the distribution.
By downloading or using this software, you accept and are bound by the terms of the license.

*/

#ifndef GENOME_H
#define GENOME_H

#include <limits>
#include <list>
#include "DnaSequence.h"
#include "Vector.h"

class Strand
{
private:
	bool _positive;

public:
	// This constructor is written like this on purpose, to avoid accidental typecasting
	Strand(void*, char p) {
		memset(this, 0, sizeof(*this));	// this is to stop valgrind from reporting unitialised memory
		if (p == '+') _positive = true;
		else if (p == '-') _positive = false;
		else {
			cerr << "Invalid strand" << endl;
			abort();
		}
	}

	static const Strand positive;
	static const Strand negative;

	bool isPositive() const { return _positive; }
	bool isNegative() const { return !_positive; }
	bool operator==(const Strand& other) const { return _positive == other._positive; }
	bool operator!=(const Strand& other) const { return _positive != other._positive; }
	Strand other() const { return _positive ? negative : positive; }
	Strand operator!() const { return _positive ? negative : positive; }
	operator char() const { return _positive ? '+' : '-'; }
};


inline ostream&
operator<<(ostream& o, Strand s)
{
	o << (s.isPositive() ? '+' : '-');
	return o;
}

typedef uint16_t ChromosomeIndex;
extern const ChromosomeIndex invalidChromosomeIndex; // = numeric_limits<ChromosomeIndex>::max();

class Locus 	// GenomicPosition + Strand
{
public:
	Locus() : _position(invalidGenomicPosition), _strand(Strand::positive)
#ifdef DEBUG
	, byte1(0), byte2(0), byte3(0)
#endif
	{}
	Locus(const GenomicPosition pos, Strand strand = Strand::positive) : _position(pos), _strand(strand)
#ifdef DEBUG
	, byte1(0), byte2(0), byte3(0)
#endif
	{}
	GenomicPosition& position() { return _position; }
	GenomicPosition position() const { return _position; }
	Locus operator+(const GenomicPosition offset) const { return Locus(_position + offset); }
	const Locus& operator+=(const GenomicPosition offset) { _position += offset; return *this; }
	Locus operator-(const GenomicPosition offset) const { assert(_position >= offset); return Locus(_position - offset); }
	const Locus& operator-=(const GenomicPosition offset) { assert(_position >= offset); _position -= offset; return *this; }
	void clear() { _position = invalidGenomicPosition; }
	bool onPositiveStrand() const { return _strand.isPositive(); }
	bool onNegativeStrand() const { return _strand.isNegative(); }
	Strand strand() const { return _strand; }
	bool isValid() const { return _position != invalidGenomicPosition; }
	bool operator<(const Locus& other) const { return _position < other._position; }
	bool operator>(const Locus& other) const { return _position > other._position; }
	bool operator==(const Locus& other) const { return _position == other._position; }
	bool operator!=(const Locus& other) const { return _position != other._position; }

protected:
	GenomicPosition _position;
	Strand _strand;

#ifdef DEBUG
private:	// this is for valgrind:
	char byte1, byte2, byte3;
	
#endif
};


inline ostream&
operator<<(ostream& o, const Locus& loc)
{
	o << (loc.position() + 1);
	return o;
}


// Note in the following a region in the negative strand also starts with the lowest coordinate, ie the end of the region is always _position + _length - 1
class GenomicRegion : public Locus
{
protected:
	GenomicPosition _length;
public:
	GenomicRegion() : _length(0) {}
	GenomicRegion(GenomicPosition pos, GenomicPosition length, Strand strand = Strand::positive) : Locus(pos, strand), _length(length) {}
	GenomicRegion(const Locus& start, GenomicPosition length) : Locus(start), _length(length) {}
	GenomicPosition start() const { return _position; }
	GenomicPosition end() const { return _position + (_length ? _length - 1 : 0); }
	Locus middle() const { return *this + _length / 2; }
	GenomicPosition& length() { return _length; }
	const GenomicPosition length() const { return _length; }
	void clear() { _length = 0; Locus::clear(); }
	Locus fivePrimeEnd() const { if (_strand.isPositive()) return *this; return  *this + (_length ? _length - 1 : 0); }
	Locus threePrimeEnd() const { if (_strand.isPositive()) return (*this + (_length ? _length - 1 : 0)); return *this; }
	bool overlaps(const GenomicRegion& other) const {
		if (_position + _length <= other._position) return false;
		if (_position >= other._position + other._length) return false;
		return true;
	}
	bool operator==(const GenomicRegion& other) const {
		return (*(Locus*)this) == (Locus&) other
		and _length == other._length;
	}
};


inline ostream&
operator<<(ostream& o, const GenomicRegion& r)
{
	const GenomicPosition pos = r.position();
	const GenomicPosition l = r.length();
	o << pos + 1 << ".." << pos + (l ? l : 1);
	return o;
}


class Genome;

class NamedGenomicRegion : public GenomicRegion
{
protected:
	String _id;

public:
	NamedGenomicRegion() {}
	NamedGenomicRegion(const String& id, GenomicPosition pos, GenomicPosition length, Strand strand = Strand::positive)
		: GenomicRegion(pos, length, strand), _id(id) {}

	static void loadBEDfile(const char* filename, const Genome&, bool firstThreeColumnsOnly, /* returned: */ list<NamedGenomicRegion>&);
	String& name() { return _id; }
	const String& name() const { return _id; }
};


inline ostream&
operator<<(ostream& o, const NamedGenomicRegion& r)
{
	o << r.name() << '=' << (const GenomicRegion&) r;
	return o;
}

class Variant;
typedef list<Variant> VariantSet;	// set of variants for a genome

// Genomic coordinates are "internal" (a single unsigned integer, unique for whole genome)
// or "external" in the form of chromosome index + chromosome position (zero-based)
// remember "GemonicPosition" and "ChromosomeIndex" are just unsigned integers
class Genome
{
public:
	Genome(const char* filename);	// loads genome from fasta file
	Genome(const Vector<DnaNamedSequence*>& chromosomes);
	Genome(const Genome&);
	~Genome() { if (_sequence) delete [] _sequence; if (_sequenceReverseComplement) delete [] _sequenceReverseComplement; }

	ChromosomeIndex numberChromosomes() const { return _chromosomeNames.size(); }
	const char* chromoSequence(ChromosomeIndex i, Strand strand = Strand::positive) const {
		if (strand.isNegative()) {
			return _sequenceReverseComplement + (_length - _chromosomeSequenceStarts[i] - _chromosomeSequenceLengths[i]);
		}
		return _sequence + _chromosomeSequenceStarts[i];
	}

	GenomicPosition length() const { return _length; }
	DnaSequence sequence(const GenomicRegion& region) const {
		if (region.onNegativeStrand()) {
			const GenomicPosition len = region.length();
			return DnaSequence(_sequenceReverseComplement + (_length - region.position() - len), len);
		}
		return DnaSequence(_sequence + region.position(), region.length());
	}
	DnaSequence sequence(GenomicPosition pos, GenomicPosition len, Strand strand = Strand::positive) const {
		if (strand.isNegative()) {
			return DnaSequence(_sequenceReverseComplement + (_length - pos - len), len);
		}
		return DnaSequence(_sequence + pos, len);
	}
	DnaSequence sequence(const Locus& loc, GenomicPosition len) const {
		if (loc.onNegativeStrand()) {
			return DnaSequence(_sequenceReverseComplement + (_length - loc.position() - len), len);
		}
		return DnaSequence(_sequence + loc.position(), len);
	}
	const char* sequence(GenomicPosition pos, Strand strand = Strand::positive) const {
		if (strand.isNegative()) {
			return _sequenceReverseComplement + (_length - pos - 1);
		}
		return _sequence + pos;
	}
	const char* sequence(const Locus& loc) const {
		if (loc.onNegativeStrand()) {
			return _sequenceReverseComplement + (_length - loc.position() - 1);
		}
		return _sequence + loc.position();
	}
	const ChromosomeIndex chromoIndex(const GenomicPosition pos) const { 
		assert(pos < length());
		ChromosomeIndex i;
		for (i = _chromosomeSequenceStarts.size()-1; pos < _chromosomeSequenceStarts[i]; i--);
		return i;
	}
	const ChromosomeIndex chromoIndex(const Locus& loc) const { return chromoIndex(loc.position()); }
	const String& chromoName(const GenomicPosition pos) const { return _chromosomeNames[chromoIndex(pos)]; }
	const String& chromoName(const Locus& loc) const { return _chromosomeNames[chromoIndex(loc)]; }

	// Converts from internal coordinates (GenomicPosition) to external (chromosome name, zero-based position):
	const String& chromoNamePos(const GenomicPosition pos, /* returned: */ GenomicPosition& chromoPos) const { 
		ChromosomeIndex i = chromoIndex(pos);
		chromoPos = pos - _chromosomeSequenceStarts[i];
		return _chromosomeNames[i];
	}
	const String& chromoNamePos(const Locus& loc, /* returned: */ GenomicPosition& chromoPos) const { 
		return chromoNamePos(loc.position(), chromoPos);
	}
	GenomicPosition chromoPos(const GenomicPosition pos) const { 
		ChromosomeIndex i = chromoIndex(pos);
		return pos - _chromosomeSequenceStarts[i];
	}
	GenomicPosition chromoPos(const Locus& l) const { return chromoPos(l.position()); }

	// Converts from external coordinates (chromosome index, zero-based position) to internal (GenomicPosition)
	GenomicPosition position(ChromosomeIndex chromoIndex, GenomicPosition chromoPos) const {
		assert(chromoIndex < _chromosomeSequenceStarts.size());
		if (chromoPos >= _chromosomeSequenceLengths[chromoIndex])
			return invalidGenomicPosition;
		return _chromosomeSequenceStarts[chromoIndex] + chromoPos;
	}
	GenomicPosition chromoStart(ChromosomeIndex ci) const {
		assert(ci < _chromosomeSequenceStarts.size());
		return _chromosomeSequenceStarts[ci];
	}
	const String& chromoName(ChromosomeIndex i) const { return _chromosomeNames[i]; }
	String toText(GenomicPosition pos, char sep = ':') const {
		GenomicPosition chromoPos = invalidGenomicPosition;
		stringstream s;
		s << chromoNamePos(pos, chromoPos) << sep;
		s << (chromoPos + 1);
		return s.str();
	}
	String toText(const Locus& loc, char sep = ':') const {
		stringstream s;
		GenomicPosition chromoPos = invalidGenomicPosition;
		s << chromoNamePos(loc, chromoPos) << sep;
		s << (chromoPos + 1);
		return s.str();
	}
	String toText(const GenomicRegion& r, char sep = ':') const { 
		stringstream s;
		GenomicPosition chromoPos = invalidGenomicPosition;
		s << chromoNamePos(r, chromoPos) << sep;
		s << (chromoPos + 1);
		if (sep == ':') s << "..";
		else s << sep;
		s << chromoPos + r.length();
		return s.str();
	}
	String toTabbedText(GenomicPosition pos) const { 
		stringstream s;
		GenomicPosition chromoPos = invalidGenomicPosition;
		s << chromoNamePos(pos, chromoPos);
		s << '\t' << (chromoPos + 1);
		return s.str();
	}
	String toTabbedText(const GenomicRegion& r) const { 
		stringstream s;
		GenomicPosition chromoPos = invalidGenomicPosition;
		s << chromoNamePos(r, chromoPos);
		s << '\t' << (chromoPos + 1) << '\t' << r.length();
		return s.str();
	}
	String toTabbedText(const NamedGenomicRegion& r) const { 
		stringstream s;
		GenomicPosition chromoPos = invalidGenomicPosition;
		s << r.name() << '\t' << chromoNamePos(r, chromoPos);
		s << '\t' << (chromoPos + 1) << '\t' << r.length();
		return s.str();
	}
	const String& filename() const { return _filename; }

	GenomicPosition chromoLength(ChromosomeIndex i) const { return _chromosomeSequenceLengths[i]; }

	uint32_t checksum() const;

	static const ChromosomeIndex SequenceNotFound = 65535;

	ChromosomeIndex chromoIndex(const String& chromoName) const {
		for (ChromosomeIndex i = 0; i < _chromosomeNames.size(); ++i) {
			if (_chromosomeNames[i] == chromoName) return i;
		}
		return SequenceNotFound;
	}

	void applyVariants(const VariantSet& v, char direction = 0);

	void output(const char* filename) const;
	void output(ostream& o) const;
	void output(ostream& o, GenomicPosition pos, char sep = ':') const {
		GenomicPosition chromoPos = invalidGenomicPosition;
		o << chromoNamePos(pos, chromoPos);
		o << sep << (chromoPos + 1);
	}
	void output(ostream& o, const Variant&) const;

	const char* sequence(Strand strand = Strand::positive) const {
		return strand.isPositive() ? _sequence : _sequenceReverseComplement;
	}
	void outputStats(ostream& o) const {
		o << "Chromosome\tStart\tLength" << endl;
		for (ChromosomeIndex i = 0; i < _chromosomeNames.size(); ++i) {
			o << _chromosomeNames[i]
			<< '\t' << _chromosomeSequenceStarts[i]
			<< '\t' << _chromosomeSequenceLengths[i]
			<< endl;
		}
	}

	void setBase(GenomicPosition pos, Base base) {
		assert(pos < _length);
		_sequence[pos] = mapBase[base];
		_sequenceReverseComplement[_length - pos - 1] = mapBase[complement(base)];
	}

	bool inSameChromosome(GenomicPosition pos1, GenomicPosition pos2) const {
		if (pos1 >= _length or pos2 >= _length) return false;
		const ChromosomeIndex i = chromoIndex(pos1);
		const GenomicPosition start = _chromosomeSequenceStarts[i];
		return (pos2 >= start and pos2 < start + _chromosomeSequenceLengths[i]);
	}

private:
	mutable String _filename;
	Vector<String> _chromosomeNames;
	Vector<GenomicPosition> _chromosomeSequenceStarts;	// not including any NNN's at both ends of chromosome
	Vector<GenomicPosition> _chromosomeSequenceLengths;
	char* _sequence;		// the whole genome concatenated
	char* _sequenceReverseComplement;
	GenomicPosition _length;

	void setFromVector(const Vector<DnaNamedSequence*>&);

//not allowed:
	Genome();
	const Genome& operator=(const Genome&);
};


inline ostream&
operator<<(ostream& o, const Genome& g)
{
	g.output(o);
	return o;
}


class Exon : public GenomicRegion
{
public:
	Exon() {}
	Exon(GenomicPosition pos, GenomicPosition length, Strand strand)
		: GenomicRegion(pos, length, strand) {}
};


class Transcript : public NamedGenomicRegion
{
public:
	Transcript() {}
	Transcript(const String& id, const GenomicRegion& cds, GenomicPosition pos, GenomicPosition length)
		: NamedGenomicRegion(id, pos, length, cds.strand()), _cds(cds) {}
	const Vector<Exon>& exons() const { return _exons; }
	const GenomicRegion& cds() const { return _cds; }
	
	static void loadUCSCtranscripts(const char* filename, const Genome&, /* return */ Vector<Transcript*> &);

private:
	GenomicRegion _cds;
	Vector<Exon> _exons;
};


#endif // GENOME_H
